



//################################################################################################
//##################                                                            ##################
//##################              Character manipulation tools                  ##################
//##################                                                            ##################
//##################                      Header  file                          ##################
//##################                                                            ##################
//################################################################################################




/*
 * Chars: High-efficiency Character manipulation tools.
 *
 * Copyright (C) 2006-2009 Yann LANDRIN-SCHWEITZER, a.k.a. Varkhan
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:  Yann LANDRIN-SCHWEITZER, a.k.a. Varkhan
 * Contact: varkhan@varkhan.net
 * Website: http://www.varkhan.net/
 * Project: http://www.varkhan.net/software/VCom
 */
#ifndef _VCOM_CHARS_H_
#define _VCOM_CHARS_H_
#include <stdarg.h>
// Regular 8-bit chars





//################################################################################################
//###
//###  Search and key lookup
//###


// Function keyncmp
// ================
// 
// Compares lexically two keys on a given number of characters.
// 
//   Args:
//         _ky1           : first key
//         _ky2           : second key
//         _len           : maximum number of characters to take into account
// 
//   Returns:
//         _ 0            : if the two keys where the same up to 'len' chars
//         _-1            : if the first key was found smaller than the second
//         _+1            : if the first key was found greater than the second

signed char vcom_keyncmp(char const * ky1, char const * ky2, unsigned long len);



// Function keysearch
// ==================
// 
// Finds the position of a particular key in an ordered array of keys.
// 
//   Args:
//         _key           : the key to look up
//         _keytab        : the arrays of keys
//         _keynum        : the size of the 'keys' table
//         _keylen        : the maximum length of keys
//         _pos           : pointer towards an integer to fill with the position
//                            of the first greater or equal key
// 
//   Returns:
//         _-1            : if the key was not found
//         otherwise the index of the given key in the table.
// 
//   Note:
//         _ The key array must be ordered in growing lexical order for this function to work.
//         _ When -1 is returned, *pos contains the index of the first greater key found,
//               that is: 0 if the key was smaller than the first in the array, keynum if
//               it was greater that the last in the array, otherwise the index of a real key.
//         _ When the key was found, *pos also holds the returned index value.

unsigned long vcom_keysearch(char const * key, char const * const * keytab, unsigned long keynum, unsigned long keylen, unsigned long * pos);





//################################################################################################
//###
//###  String formatting
//###




// Function strnfmt
// ================
// 
// Formats a string and a variadic set of arguments into a buffer.
// 
//   Args:
//         _str           : a character buffer where the formated string will be written
//         _len           : the maximum number of characters to write to the buffer
//         _fmt           : the format string
//         _ap            : a variadic set of arguments
// 
//   Returns:
//         _-1            : if the format string is invalid
//         _-2            : if an argument in the variadic set is invalid
//         otherwise the number of characters written to the buffer
// 
//   Note:
//          All characters from the format are added to the buffer, except those within an escape sequence.
//          Escape sequences start with a '%' character, can specify several optional parameters, and end
//          with a format type character. Each escape sequence consumes one argument from the array, and
//          transforms that argument into a format item, appended to the buffer, according to the type and
//          parameters of the escape sequence.
//          .
//          Accepted format types are:
//          _ % : outputs a literal '%' character, *without consuming an argument*
//          _ c : formats the argument as a single character having the integer value of the argument
//          _ s : formats the argument as a string (argument must be a pointer to an array of characters, terminated by a nul byte)
//          _ d : formats the argument as a signed decimal integer
//          _ u : formats the argument as an unsigned decimal integer
//          _ x : formats the argument as an hexadecimal integer, using lower case letters
//          _ X : formats the argument as an hexadecimal integer, using upper case letters
//          .
//          Parameters can be, in order, any of the optional following:
//          _ < or > : an orientation specifier, defining the alignment direction of the formatting: respectively left, or right (default: right)
//          _ |      : a flag indicating that if the formatting results in more characters than the specified size, it should be truncated (default: none)
//          _ + or - : specifies that signed formatting of a positive number should respectively include or omit a leading '+' character (default: omit)
//          _ a pad char (any character except digits from 1 to 9)
//          _ a decimal number : indicating the size (number of characters) of the formatted result
//          _ . (a dot) followed by a decimal number : indicating, for unsigned decimal integers: the radix (default: 10),
//              for hexadecimal integers: the shift (the number of bits that each digit represents, i.e. the base-2 logarithm of the radix)

unsigned long vcom_strnfmt_v(char * str, unsigned long len, char const * fmt, va_list ap);

unsigned long vcom_strnfmt(char * str, unsigned long len, char const * fmt, ...);



// Function wtrnfmt
// ================
// 
// Formats a string and a variadic set of arguments for a writer.
// 
//   Args:
//         _wtr           : a writer function pointer: _Chars_size_t (*)(void * ptr, const void * buf, size_t len)
//         _ptr           : the 'handle' reference pointer for the writer function
//         _fmt           : the format string
//         _ap            : a variadic set of arguments
// 
//   Returns:
//         _-1            : if the format string is invalid
//         _-2            : if an argument in the variadic set is invalid
//         otherwise the number of characters written
// 
//   Note:
//          All characters from the format are written unchanged, except those within an escape sequence.
//          Escape sequences start with a '%' character, can specify several optional parameters, and end
//          with a format type character. Each escape sequence consumes one argument from the array, and
//          transforms that argument into a format item, appended to the buffer, according to the type and
//          parameters of the escape sequence.
//          .
//          Accepted format types are:
//          _ % : outputs a literal '%' character, *without consuming an argument*
//          _ c : formats the argument as a single character having the integer value of the argument
//          _ s : formats the argument as a string (argument must be a pointer to an array of characters, terminated by a nul byte)
//          _ d : formats the argument as a signed decimal integer
//          _ u : formats the argument as an unsigned decimal integer
//          _ x : formats the argument as an hexadecimal integer, using lower case letters
//          _ X : formats the argument as an hexadecimal integer, using upper case letters
//          .
//          Parameters can be, in order, any of the optional following:
//          _ < or > : an orientation specifier, defining the alignment direction of the formatting: respectively left, or right (default: right)
//          _ |      : a flag indicating that if the formatting results in more characters than the specified size, it should be truncated (default: none)
//          _ + or - : specifies that signed formatting of a positive number should respectively include or omit a leading '+' character (default: omit)
//          _ a pad char (any character except digits from 1 to 9)
//          _ a decimal number : indicating the size (number of characters) of the formatted result
//          _ . (a dot) followed by a decimal number : indicating, for unsigned decimal integers: the radix (default: 10),
//              for hexadecimal integers: the shift (the number of bits that each digit represents, i.e. the base-2 logarithm of the radix)

unsigned long vcom_wtrnfmt_v(unsigned long (* wtr)(void * ptr, void const * buf, size_t len), void * ptr, char const * fmt, va_list ap);

unsigned long vcom_wtrnfmt(unsigned long (* wtr)(void * ptr, void const * buf, size_t len), void * ptr, char const * fmt, ...);




//###
//###  END
//###
//################################################################################################
// Multibyte chars
#include <wchar.h>
//################################################################################################
//###
//###  Search and key lookup
//###
// Function keyncmp
// ================
// 
// Compares lexically two keys on a given number of characters.
// 
//   Args:
//         _ky1           : first key
//         _ky2           : second key
//         _len           : maximum number of characters to take into account
// 
//   Returns:
//         _ 0            : if the two keys where the same up to 'len' chars
//         _-1            : if the first key was found smaller than the second
//         _+1            : if the first key was found greater than the second
signed char vcom_wkeyncmp(wchar_t const * ky1, wchar_t const * ky2, unsigned long len);
// Function keysearch
// ==================
// 
// Finds the position of a particular key in an ordered array of keys.
// 
//   Args:
//         _key           : the key to look up
//         _keytab        : the arrays of keys
//         _keynum        : the size of the 'keys' table
//         _keylen        : the maximum length of keys
//         _pos           : pointer towards an integer to fill with the position
//                            of the first greater or equal key
// 
//   Returns:
//         _-1            : if the key was not found
//         otherwise the index of the given key in the table.
// 
//   Note:
//         _ The key array must be ordered in growing lexical order for this function to work.
//         _ When -1 is returned, *pos contains the index of the first greater key found,
//               that is: 0 if the key was smaller than the first in the array, keynum if
//               it was greater that the last in the array, otherwise the index of a real key.
//         _ When the key was found, *pos also holds the returned index value.
unsigned long vcom_wkeysearch(wchar_t const * key, wchar_t const * const * keytab, unsigned long keynum, unsigned long keylen, unsigned long * pos);
//################################################################################################
//###
//###  String formatting
//###
// Function strnfmt
// ================
// 
// Formats a string and a variadic set of arguments into a buffer.
// 
//   Args:
//         _str           : a character buffer where the formated string will be written
//         _len           : the maximum number of characters to write to the buffer
//         _fmt           : the format string
//         _ap            : a variadic set of arguments
// 
//   Returns:
//         _-1            : if the format string is invalid
//         _-2            : if an argument in the variadic set is invalid
//         otherwise the number of characters written to the buffer
// 
//   Note:
//          All characters from the format are added to the buffer, except those within an escape sequence.
//          Escape sequences start with a '%' character, can specify several optional parameters, and end
//          with a format type character. Each escape sequence consumes one argument from the array, and
//          transforms that argument into a format item, appended to the buffer, according to the type and
//          parameters of the escape sequence.
//          .
//          Accepted format types are:
//          _ % : outputs a literal '%' character, *without consuming an argument*
//          _ c : formats the argument as a single character having the integer value of the argument
//          _ s : formats the argument as a string (argument must be a pointer to an array of characters, terminated by a nul byte)
//          _ d : formats the argument as a signed decimal integer
//          _ u : formats the argument as an unsigned decimal integer
//          _ x : formats the argument as an hexadecimal integer, using lower case letters
//          _ X : formats the argument as an hexadecimal integer, using upper case letters
//          .
//          Parameters can be, in order, any of the optional following:
//          _ < or > : an orientation specifier, defining the alignment direction of the formatting: respectively left, or right (default: right)
//          _ |      : a flag indicating that if the formatting results in more characters than the specified size, it should be truncated (default: none)
//          _ + or - : specifies that signed formatting of a positive number should respectively include or omit a leading '+' character (default: omit)
//          _ a pad char (any character except digits from 1 to 9)
//          _ a decimal number : indicating the size (number of characters) of the formatted result
//          _ . (a dot) followed by a decimal number : indicating, for unsigned decimal integers: the radix (default: 10),
//              for hexadecimal integers: the shift (the number of bits that each digit represents, i.e. the base-2 logarithm of the radix)
unsigned long vcom_wstrnfmt_v(wchar_t * str, unsigned long len, wchar_t const * fmt, va_list ap);
unsigned long vcom_wstrnfmt(wchar_t * str, unsigned long len, wchar_t const * fmt, ...);
// Function wtrnfmt
// ================
// 
// Formats a string and a variadic set of arguments for a writer.
// 
//   Args:
//         _wtr           : a writer function pointer: _Chars_size_t (*)(void * ptr, const void * buf, size_t len)
//         _ptr           : the 'handle' reference pointer for the writer function
//         _fmt           : the format string
//         _ap            : a variadic set of arguments
// 
//   Returns:
//         _-1            : if the format string is invalid
//         _-2            : if an argument in the variadic set is invalid
//         otherwise the number of characters written
// 
//   Note:
//          All characters from the format are written unchanged, except those within an escape sequence.
//          Escape sequences start with a '%' character, can specify several optional parameters, and end
//          with a format type character. Each escape sequence consumes one argument from the array, and
//          transforms that argument into a format item, appended to the buffer, according to the type and
//          parameters of the escape sequence.
//          .
//          Accepted format types are:
//          _ % : outputs a literal '%' character, *without consuming an argument*
//          _ c : formats the argument as a single character having the integer value of the argument
//          _ s : formats the argument as a string (argument must be a pointer to an array of characters, terminated by a nul byte)
//          _ d : formats the argument as a signed decimal integer
//          _ u : formats the argument as an unsigned decimal integer
//          _ x : formats the argument as an hexadecimal integer, using lower case letters
//          _ X : formats the argument as an hexadecimal integer, using upper case letters
//          .
//          Parameters can be, in order, any of the optional following:
//          _ < or > : an orientation specifier, defining the alignment direction of the formatting: respectively left, or right (default: right)
//          _ |      : a flag indicating that if the formatting results in more characters than the specified size, it should be truncated (default: none)
//          _ + or - : specifies that signed formatting of a positive number should respectively include or omit a leading '+' character (default: omit)
//          _ a pad char (any character except digits from 1 to 9)
//          _ a decimal number : indicating the size (number of characters) of the formatted result
//          _ . (a dot) followed by a decimal number : indicating, for unsigned decimal integers: the radix (default: 10),
//              for hexadecimal integers: the shift (the number of bits that each digit represents, i.e. the base-2 logarithm of the radix)
unsigned long vcom_wwtrnfmt_v(unsigned long (* wtr)(void * ptr, void const * buf, size_t len), void * ptr, wchar_t const * fmt, va_list ap);
unsigned long vcom_wwtrnfmt(unsigned long (* wtr)(void * ptr, void const * buf, size_t len), void * ptr, wchar_t const * fmt, ...);
//###
//###  END
//###
//################################################################################################
#endif // _VCOM_CHARS_H_
