
/*################################################################################################
 *##################                                                            ##################
 *##################              Character manipulation tools                  ##################
 *##################                                                            ##################
 *##################                                                            ##################
 *##################                     Implementation                         ##################
 *##################                                                            ##################
 *################################################################################################
 */

/*
 * Chars: High-efficiency Character manipulation tools.
 *
 * Copyright (C) 2006-2009 Yann LANDRIN-SCHWEITZER, a.k.a. Varkhan
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:  Yann LANDRIN-SCHWEITZER, a.k.a. Varkhan
 * Contact: varkhan@varkhan.net
 * Website: http://www.varkhan.net/
 * Project: http://www.varkhan.net/software/VCom
 */



#include <unistd.h>
#include <stdarg.h>

#include "_Chars_.h"



//################################################################################################
//###
//###  Search and key lookup
//###


// Function keyncmp
// ================
// 
// Compares lexically two keys on a given number of characters.
// 
//   Args:
//         _ky1           : first key
//         _ky2           : second key
//         _len           : maximum number of characters to take into account
// 
//   Returns:
//         _ 0            : if the two keys where the same up to 'len' chars
//         _-1            : if the first key was found smaller than the second
//         _+1            : if the first key was found greater than the second

_Chars_sign_t _Chars_m(keyncmp)(_Chars_char_t const * ky1, _Chars_char_t const * ky2, _Chars_size_t len) {
	register _Chars_size_t i = 0;
	while(i<len) {
		if(ky1[i]<ky2[i]) return -1;
		if(ky1[i]>ky2[i]) return +1;
		if(ky1[i]==0) break;
		i ++;
	}
	return 0;
}

// Function keyhash
// ================
// 
// Compares lexically two keys on a given number of characters.
// 
//   Args:
//         _key           : the key to compute a mixing value of
//         _len           : maximum number of characters to take into account
// 
//   Returns:
//         a mixing value for the key

_Chars_sign_t _Chars_m(keyhash)(_Chars_char_t const * key, _Chars_size_t len) {
	register _Chars_size_t i = 0;
	register _Chars_size_t x;
	switch(sizeof(_Chars_char_t)) {
		case 1: x = 3; break;
		case 2: x = 4; break;
		case 4: x = 5; break;
		case 8: x = 6; break;
	}
	register _Chars_size_t p = ( sizeof(_Chars_size_t) / sizeof(_Chars_char_t) ) - 1;
	register _Chars_size_t m = 0;
	while(i<len) {
		m ^= key[i] << ( (i & p) << x );
		i ++;
	}
	return m;
}

// Function keysearch
// ==================
// 
// Finds the position of a particular key in an ordered array of keys.
// 
//   Args:
//         _key           : the key to look up
//         _keytab        : the arrays of keys
//         _keynum        : the size of the 'keys' table
//         _keylen        : the maximum length of keys
//         _pos           : pointer towards an integer to fill with the position
//                            of the first greater or equal key
// 
//   Returns:
//         _-1            : if the key was not found
//         otherwise the index of the given key in the table.
//
//   Note:
//         _ The key array must be ordered in growing lexical order for this function to work.
//         _ When -1 is returned, *pos contains the index of the first greater key found,
//               that is: 0 if the key was smaller than the first in the array, keynum if
//               it was greater that the last in the array, otherwise the index of a real key.
//         _ When the key was found, *pos also holds the returned index value.

_Chars_size_t _Chars_m(keysearch)(_Chars_char_t const * key, _Chars_char_t const * const * keytab, _Chars_size_t keynum, _Chars_size_t keylen, _Chars_size_t * pos) {
	register _Chars_size_t min, max, med;
	if(keynum==0) {
		if(pos!=NULL) *pos = 0;
		return -1;
	}
	min = 0;
	max = keynum;
	switch( _Chars_m(keyncmp)(key,keytab[min],keylen)) {
		case -1:
			if(pos!=NULL) *pos = min;
			return -1;
		case +1:
			break;
		case 0:
			if(pos!=NULL) *pos = min;
			return min;
		default:
			return -1;
	}
	switch( _Chars_m(keyncmp)(key,keytab[max-1],keylen)) {
		case -1:
			break;
		case +1:
			if(pos!=NULL) *pos = max;	
			return -1;
		case 0:
			if(pos!=NULL) *pos = max-1;
			return max-1;
		default:
			return -1;
	}
	while(min<max) {
		med = (min+max-1)>>1;
		switch( _Chars_m(keyncmp)(key,keytab[med],keylen)) {
			case -1:
				max = med;
				break;
			case +1:
				min = med+1;
				break;
			case 0:
				if(pos!=NULL) *pos = med;
				return med;
			default:
				return -1;
		}
	}
	// At this point: min == max
	if(pos!=NULL) *pos = min;
	// Ensure we haven't some pathological case such as keynum == 1
	if( _Chars_m(keyncmp)(key,keytab[min],keylen)==0) return min;
	return -1;
}





//################################################################################################
//###
//###  String formatting
//###



#define CHARS_FMTERR 0x01
#define CHARS_ARGERR 0x02
#define CHARS_ARGMIS 0x04


#define CHARS_STRBUF   64
#define CHARS_TMPBUF 1024


// Function vstrnfmt
// =================
// 
// Formats a string and a variadic set of arguments into a buffer.
// 
//   Args:
//         _str           : a character buffer where the formated string will be written
//         _len           : the maximum number of characters to write to the buffer
//         _fmt           : the format string, defining how the arguments should be written to the buffer
//         _ap            : a variadic set of arguments
// 
//   Returns:
//         _-1            : if the format string is invalid
//         _-2            : if an argument in the variadic set is invalid
//         otherwise the number of characters written to the buffer
//
//   Note:
//          All characters from the format are added unchanged to the buffer, except those within an escape sequence.
//          Escape sequences start with a '%' character, can specify several optional parameters, and end
//          with a format type character. Each escape sequence consumes one argument from the array, and
//          transforms that argument into a format item, appended to the buffer, according to the type and
//          parameters of the escape sequence.
//          .
//          Accepted format types are:
//          _ % : outputs a literal '%' character, *without consuming an argument*
//          _ c : formats the argument as a single character having the integer value of the argument
//          _ s : formats the argument as a string (argument must be a pointer to an array of characters, terminated by a nul byte)
//          _ d : formats the argument as a signed decimal integer
//          _ u : formats the argument as an unsigned decimal integer
//          _ x : formats the argument as an hexadecimal integer, using lower case letters
//          _ X : formats the argument as an hexadecimal integer, using upper case letters
//          .
//          Parameters can be, in order, any of the optional following:
//          _ < or > : an orientation specifier, defining the alignment direction of the formatting: respectively left, or right (default: right)
//          _ |      : a flag indicating that if the formatting results in more characters than the specified size, it should be truncated (default: none)
//          _ + or - : specifies that signed formatting of a positive number should respectively include or omit a leading '+' character (default: omit)
//          _ a pad char (any character except digits from 1 to 9)
//          _ a decimal number : indicating the size (number of characters) of the formatted result
//          _ . (a dot) followed by a decimal number : indicating, for unsigned decimal integers: the radix (default: 10),
//              for hexadecimal integers: the shift (the number of bits that each digit represents, i.e. the base-2 logarithm of the radix)

_Chars_size_t _Chars_v(strnfmt)(_Chars_char_t * str, _Chars_size_t len, _Chars_char_t const * fmt, va_list ap) {
	register _Chars_char_t chr;
	register _Chars_size_t pos = 0;
	register _Chars_size_t out = 0;
	_Chars_char_t buf[CHARS_STRBUF];
	while( (chr=fmt[pos++]) != '\0' ) {
		// If not escape, simply add char
		if(chr!='%') {
			if(out<len) str[out++] = chr;
			continue;
		}
		chr = fmt[pos++];
		if(chr=='\0') return -CHARS_FMTERR;
		// ### Parse format parameters ####################################################
		_Chars_bool_t sign = 0;	// Whether a + sign should prefix positive numbers
		_Chars_bool_t left = 0;	// Whether item should be aligned left instead of right
		_Chars_bool_t trnc = 0;	// Whether oversized items should be truncated
		if(chr=='<') {
			left = 1;
			chr = fmt[pos++];
			if(chr=='\0') return -CHARS_FMTERR;
		}
		else if(chr=='>') {
			left = 0;
			chr = fmt[pos++];
			if(chr=='\0') return -CHARS_FMTERR;
		}
		if(chr=='|') {
			trnc = 1;
			chr = fmt[pos++];
			if(chr=='\0') return -CHARS_FMTERR;
		}
		if(chr=='+') {
			sign = 1;
			chr = fmt[pos++];
			if(chr=='\0') return -CHARS_FMTERR;
		}
		else if(chr=='-') {
			sign = 0;
			chr = fmt[pos++];
			if(chr=='\0') return -CHARS_FMTERR;
		}
		_Chars_char_t pad = '\0';
		// --- Look for a pad char --------------------------------------------------------
		if( ((chr<='0')||(chr> '9'))
				&& ((chr< 'a')||(chr> 'z'))
				&& ((chr< 'A')||(chr> 'Z'))
				&& (chr!='%') ) {
			pad = chr;
			chr = fmt[pos++];
			if(chr=='\0') return -CHARS_FMTERR;
		}
		// --- Look for display width -----------------------------------------------------
		_Chars_size_t dsp = 0;
		if(chr=='*') {
			dsp = (_Chars_size_t) va_arg(ap, int);
			chr = fmt[pos++];
			if(chr=='\0') return -CHARS_FMTERR;
		}
		else if( (chr>'0')&&(chr<='9') ) {
			while( (chr>='0')&&(chr<='9') ) {
				dsp *= 10;
				dsp += chr-'0';
				chr = fmt[pos++];
				if(chr=='\0') return -CHARS_FMTERR;
			}
			if(dsp<0) {
				left = !left;
				dsp = -dsp;
			}
		}
		// --- Look for display precision / radix / shift ---------------------------------
		_Chars_size_t pre = 0;
		if(chr=='.') {
			chr = fmt[pos++];
			if(chr=='\0') return -CHARS_FMTERR;
			if(chr=='*') {
				pre = (_Chars_size_t) va_arg(ap, int);
				chr = fmt[pos++];
				if(chr=='\0') return -CHARS_FMTERR;
			}
			else if( (chr>'0')&&(chr<='9') ) {
				while( (chr>='0')&&(chr<='9') ) {
					pre *= 10;
					pre += chr-'0';
					chr = fmt[pos++];
					if(chr=='\0') return -CHARS_FMTERR;
				}
				if(pre<0) return -CHARS_ARGERR;
			}
		}
		// ### Switch on format type ######################################################
		switch(chr) {
			case '%':			// --- Literal % --------------------------
				if(out<len) str[out++] = chr;
				break;
			case 'c':			// --- Char format ------------------------
				chr = (_Chars_char_t) va_arg(ap, int);
				if(out<len) str[out++] = chr;
				break;
			case 's': {			// --- String format ----------------------
				_Chars_char_t const * s = (_Chars_char_t *) va_arg(ap, void *);
				if(s==NULL) return -CHARS_ARGERR;
				// Pad with ' ' by default
				if(pad=='\0') pad = ' ';
				if(left) {		// Let align
					_Chars_size_t i = 0;
					if(trnc) while(i<dsp && (chr=s[i++])!='\0' && out<len) str[out++] = chr;
					else while((chr=s[i++])!='\0' && out<len) str[out++] = chr;
					while(i++<dsp && out<len) str[out++] = pad;
				}
				else {			// Right align
					_Chars_size_t i = 0;
					while(s[i++]!='\0');
					// Pad display chars until i>=dsp
					while(i<dsp-- && out<len) str[out++] = pad;
					// Either we padded enough, or dsp was smaller that i
					if(trnc) i -= dsp;	// Write all chars from len-dsp to len
					else i = 0;		// Write all chars from 0 to len
					while((chr=s[i++])!='\0' && out<len) str[out++] = chr;
				}
			}	break;
			case 'd': {			// --- Decimal signed integer -------------
				const signed long v = (signed long) va_arg(ap, int);
				unsigned long u = (v<0)?-v:v;
				_Chars_size_t i = 0;
				// Pad with ' ' by default
				if(pad=='\0') pad = ' ';
				// Base is 10 by default
				if(pre==0) pre = 10;
				while(i<CHARS_STRBUF && u!=0) {
					buf[i++] = '0' + u%pre;
					u /= pre;
				}
				if(dsp==0) {		// Fast version if no format width
					while(i>0 && out<len) str[out++] = buf[--i];
				}
				else if(pad=='0') {	// '0' implies right align
					// Append sign
					if(v<0 && out<len) { str[out++] = '-'; dsp --; }
					else if(v>0 && sign && out<len) { str[out++] = '+'; dsp --; }
					// Append padding
					while(dsp>i && out<len) { str[out++] = pad; dsp --; }
					// Append digits
					while(i>0 && out<len) str[out++] = buf[--i];
				}
				else if(left) {		// Left align
					// Append sign
					if(v<0 && out<len) { str[out++] = '-'; dsp --; }
					else if(v>0 && sign && out<len) { str[out++] = '+'; dsp --; }
					// Append digits
					dsp -= i;
					while(i>0 && out<len) str[out++] = buf[--i];
					// Append padding
					while(dsp>0 && out<len) { str[out++] = pad; dsp --; }
				}
				else {			// Right align
					// Reserve space for sign
					if(v<0) dsp --;
					else if(v>0 && sign) dsp --;
					// Append padding
					while(dsp>i && out<len) { str[out++] = pad; dsp --; }
					// Append sign
					if(v<0 && out<len) str[out++] = '-';
					else if(v>0 && sign && out<len) str[out++] = '+';
					// Append digits
					while(i>0 && out<len) str[out++] = buf[--i];
				}
				  } break;
			case 'u': {			// --- Decimal unsigned integer -----------
				unsigned long u = (unsigned long) va_arg(ap, int);
				_Chars_size_t i = 0;
				// Pad with ' ' by default
				if(pad=='\0') pad = ' ';
				// Base is 10 by default
				if(pre==0) pre = 10;
				while(i<CHARS_STRBUF && u!=0) {
					buf[i++] = '0' + u%pre;
					u /= pre;
				}
				if(dsp==0) {		// Fast version if no format width
					while(i>0 && out<len) str[out++] = buf[--i];
				}
				else if(left) {		// Left align
					// Append digits
					dsp -= i;
					while(i>0 && out<len) str[out++] = buf[--i];
					// Append padding
					while(dsp>0 && out<len) { str[out++] = pad; dsp --; }
				}
				else {			// Right align
					// Append padding
					while(dsp>i && out<len) { str[out++] = pad; dsp --; }
					// Append digits
					while(i>0 && out<len) str[out++] = buf[--i];
				}
			}	break;
			case 'x': case 'X': {		// --- Hexadecimal unsigned integer -------
				unsigned long u = (unsigned long) va_arg(ap, int);
				_Chars_size_t i = 0;
				// Pad with ' ' by default
				if(pad=='\0') pad = ' ';
				// Shift is 4 by default
				if(pre==0) pre = 4;
				if(pre>=6) pre = 6;
				unsigned char msk = (1<<pre)-1;
				if(chr=='X') chr = 'A';
				else chr = 'a';
				while(i<CHARS_STRBUF && u!=0) {
					unsigned char d = u&msk;
					if(d<10) buf[i++] = '0' + d;
					else buf[i++] = chr + (d-10);
					u >>= pre;
				}
				if(left) {
					// Append digits
					while(i>0 && dsp>0 && out<len) { str[out++] = buf[--i]; dsp --; }
					// If not truncating, write overflowing lower order digits
					if(!trnc) while(i>0 && out<len) str[out++] = buf[--i];
					// Append padding
					while(dsp>0 && out<len) { str[out++] = pad; dsp --; }
				}
				else {
					// Append padding
					while(dsp>i && out<len) { str[out++] = pad; dsp --; }
					// If truncating and overflow, keep only lower order digits
					if(trnc) i = dsp;
					// Append digits
					while(i>0 && out<len) str[out++] = buf[--i];
				}
			}	break;
			default: {			// --- Other format types: ignored --------
			}
		} // End format character switch
	} // End Item loop
	if(out<len) str[out++] = '\0';			// - Terminate C-string -------------------
	// ### Return number of characters written ################################################
	return out;
}


_Chars_size_t _Chars_m(strnfmt)(_Chars_char_t * str, _Chars_size_t len, _Chars_char_t const * fmt, ...) {
	va_list ap;
	_Chars_size_t res;
	va_start(ap,fmt);
	res = _Chars_v(strnfmt)(str,len,fmt,ap);
	va_end(ap);
	return res;
}


// Function wtrnfmt
// ================
// 
// Formats a string and a variadic set of arguments into a buffer.
// 
//   Args:
//         _wtr           : a writer function pointer: ssize_t (*)(void * ptr, const void * buf, size_t len)
//         _ptr           : the 'handle' reference pointer for the writer function
//         _fmt           : the format string, defining how the arguments should be written to the buffer
//         _ap            : a variadic set of arguments
// 
//   Returns:
//         _-1            : if the format string is invalid
//         _-2            : if an argument in the variadic set is invalid
//         otherwise the number of characters written
// 
//   Note:
//          All characters from the format are written unchanged, except those within an escape sequence.
//          Escape sequences start with a '%' character, can specify several optional parameters, and end
//          with a format type character. Each escape sequence consumes one argument from the array, and
//          transforms that argument into a format item, appended to the buffer, according to the type and
//          parameters of the escape sequence.
//          .
//          Accepted format types are:
//          Accepted format types are:
//          _ % : outputs a literal '%' character, *without consuming an argument*
//          _ c : formats the argument as a single character
//          _ s : formats the argument as a character sequence
//          _ d : formats the argument as a signed decimal integer
//          _ u : formats the argument as an unsigned decimal integer
//          _ x : formats the argument as an hexadecimal integer, using lower case letters
//          _ X : formats the argument as an hexadecimal integer, using upper case letters
//          .
//          Parameters in order, any of the optional following:
//          _ < or > : an orientation specifier, defining the alignment direction of the formatting: respectively left, or right (default: right)
//          _ |      : a flag indicating that if the formatting results in more characters than the specified size, it should be truncated (default: none)
//          _ + or - : specifies that signed formatting of a positive number should respectively include or omit a leading '+' character (default: omit)
//          _ a pad char (any character except digits from 1 to 9)
//          _ a decimal number : indicating the size (number of characters) of the formatted result
//          _ . (a dot) followed by a decimal number : indicating, for unsigned decimal integers: the radix (default: 10),
//              for hexadecimal integers: the shift (the number of bits that each digit represents, i.e. the base-2 logarithm of the radix)

_Chars_size_t _Chars_v(wtrnfmt)(_Chars_size_t (* wtr)(void * ptr, void const * buf, size_t len), void * ptr, _Chars_char_t const * fmt, va_list ap) {
	register _Chars_char_t chr;
	register _Chars_size_t pos = 0;
	register _Chars_size_t out = 0;
	_Chars_size_t len = 0;
	_Chars_size_t w = 0;
	_Chars_char_t buf[CHARS_STRBUF];
	_Chars_char_t tmp[CHARS_TMPBUF];
	while( (chr=fmt[pos++]) != '\0' ) {
		// If not escape, simply add all chars up to next escape
		if(chr!='%') {
			_Chars_size_t i = 0;
			while( (chr=fmt[pos+i++]) != '\0' ) { if(chr=='%') break; }
			w = wtr(ptr,fmt+pos,i);
			if(w<0) return w;
			len += w;
			if(w<i) return len;
			continue;
		}
		chr = fmt[pos++];
		if(chr=='\0') return -CHARS_FMTERR;
		// ### Parse format parameters ####################################################
		_Chars_bool_t sign = 0;	// Whether a plus sign should be appended to positive numeric items
		_Chars_bool_t left = 0;	// Whether item should be aligned left instead of right
		_Chars_bool_t trnc = 0;	// Whether oversized items should be truncated
		if(chr=='<') {
			left = 1;
			chr = fmt[pos++];
			if(chr=='\0') return -CHARS_FMTERR;
		}
		else if(chr=='>') {
			left = 0;
			chr = fmt[pos++];
			if(chr=='\0') return -CHARS_FMTERR;
		}
		if(chr=='|') {
			trnc = 1;
			chr = fmt[pos++];
			if(chr=='\0') return -CHARS_FMTERR;
		}
		if(chr=='+') {
			sign = 1;
			chr = fmt[pos++];
			if(chr=='\0') return -CHARS_FMTERR;
		}
		else if(chr=='-') {
			sign = 0;
			chr = fmt[pos++];
			if(chr=='\0') return -CHARS_FMTERR;
		}
		_Chars_char_t pad = '\0';
		// --- Look for a pad char --------------------------------------------------------
		if( ((chr<='0')||(chr> '9'))
				&& ((chr< 'a')||(chr> 'z'))
				&& ((chr< 'A')||(chr> 'Z'))
				&& (chr!='%') ) {
			pad = chr;
			chr = fmt[pos++];
			if(chr=='\0') return -CHARS_FMTERR;
		}
		// --- Look for display width -----------------------------------------------------
		_Chars_size_t dsp = 0;
		if(chr=='*') {
			dsp = (_Chars_size_t) va_arg(ap, int);
			chr = fmt[pos++];
			if(chr=='\0') return -CHARS_FMTERR;
		}
		else if( (chr>'0')&&(chr<='9') ) {
			while( (chr>='0')&&(chr<='9') ) {
				dsp *= 10;
				dsp += chr-'0';
				chr = fmt[pos++];
				if(chr=='\0') return -CHARS_FMTERR;
			}
			if(dsp<0) {
				left = !left;
				dsp = -dsp;
			}
		}
		// --- Look for display precision / radix / shift ---------------------------------
		_Chars_size_t pre = 0;
		if(chr=='.') {
			chr = fmt[pos++];
			if(chr=='\0') return -CHARS_FMTERR;
			if(chr=='*') {
				pre = (_Chars_size_t) va_arg(ap, int);
				chr = fmt[pos++];
				if(chr=='\0') return -CHARS_FMTERR;
			}
			else if( (chr>'0')&&(chr<='9') ) {
				while( (chr>='0')&&(chr<='9') ) {
					pre *= 10;
					pre += chr-'0';
					chr = fmt[pos++];
					if(chr=='\0') return -CHARS_FMTERR;
				}
				if(pre<0) return -CHARS_ARGERR;
			}
		}
		// ### Switch on format type ######################################################
		switch(chr) {
			case '%':			// --- Literal % --------------------------
				pad = chr;
				w = wtr(ptr,&pad,1);
				if(w<0) return w;
				len += w;
				if(w<1) return len;
				break;
			case 'c':			// --- Char format ------------------------
				pad = (_Chars_char_t) va_arg(ap, int);
				w = wtr(ptr,&pad,1);
				if(w<0) return w;
				len += w;
				if(w<1) return len;
				break;
			case 's': {			// --- String format ----------------------
				_Chars_char_t const * s = (_Chars_char_t *) va_arg(ap, void *);
				_Chars_size_t i = 0;
				if(s==NULL) return -CHARS_ARGERR;
				while(s[i++]!='\0');
				// Pad with ' ' by default
				if(pad=='\0') pad = ' ';
				if(left) {		// Let align
					if(trnc && i>dsp) {
						w = wtr(ptr,s,dsp);
						if(w<0) return w;
						len += w;
						if(w<dsp) return len;
					}
					else {
						w = wtr(ptr,s,i);
						if(w<0) return w;
						len += w;
						if(w<i) return len;
						// Bufferize the padding until i>=dsp
						dsp -= i;
						while(dsp>0) {
							_Chars_size_t j = 0;
							while(j<CHARS_TMPBUF && j<dsp) tmp[j++] = pad;
							w = wtr(ptr,tmp,j);
							if(w<0) return w;
							len += w;
							if(w<j) return len;
							dsp -= j;
						}
					}
				}
				else {			// Right align
					if(trnc && i>dsp) {
						w = wtr(ptr,s+i-dsp,dsp);
						if(w<0) return w;
						len += w;
						if(w<dsp) return len;
					}
					else {
						// Bufferize the padding until i>=dsp
						dsp -= i;
						while(dsp>0) {
							_Chars_size_t j = 0;
							while(j<CHARS_TMPBUF && j<dsp) tmp[j++] = pad;
							w = wtr(ptr,tmp,j);
							if(w<0) return w;
							len += w;
							if(w<j) return len;
							dsp -= j;
						}
						w = wtr(ptr,s,i);
						if(w<0) return w;
						len += w;
						if(w<i) return len;
					}
				}
			}	break;
			case 'd': {			// --- Decimal signed integer -------------
				const signed long v = (signed long) va_arg(ap, int);
				unsigned long u = (v<0)?-v:v;
				_Chars_size_t i = 0;
				// Pad with ' ' by default
				if(pad=='\0') pad = ' ';
				// Base is 10 by default
				if(pre==0) pre = 10;
				i = CHARS_STRBUF;
				while(i>0 && u!=0) {
					buf[--i] = '0' + u%pre;
					u /= pre;
				}
				if(dsp==0) {		// Fast version if no format width
					w = wtr(ptr,buf+i,CHARS_STRBUF-i);
					if(w<0) return w;
					len += w;
					if(w<(CHARS_STRBUF-i)) return len;
				}
				else if(pad=='0') {	// '0' implies right align
					// Append sign
					if(v<0) { 
						_Chars_char_t sgn = '-';
						w = wtr(ptr,&sgn,1);
						if(w<0) return w;
						len += w;
						if(w<1) return len;
						dsp --;
					}
					else if(v>0 && sign && out<len) {
						_Chars_char_t sgn = '+';
						w = wtr(ptr,&sgn,1);
						if(w<0) return w;
						len += w;
						if(w<1) return len;
						dsp --;
					}
					// Append padding
					dsp -= CHARS_STRBUF-i;
					while(dsp>0) {
						_Chars_size_t j = 0;
						while(j<CHARS_TMPBUF && j<dsp) tmp[j++] = pad;
						w = wtr(ptr,tmp,j);
						if(w<0) return w;
						len += w;
						if(w<j) return len;
						dsp -= j;
					}
					// Append digits
					w = wtr(ptr,buf+i,CHARS_STRBUF-i);
					if(w<0) return w;
					len += w;
					if(w<(CHARS_STRBUF-i)) return len;
				}
				else if(left) {		// Left align
					// Append sign
					if(v<0) { 
						_Chars_char_t sgn = '-';
						w = wtr(ptr,&sgn,1);
						if(w<0) return w;
						len += w;
						if(w<1) return len;
						dsp --;
					}
					else if(v>0 && sign && out<len) {
						_Chars_char_t sgn = '+';
						w = wtr(ptr,&sgn,1);
						if(w<0) return w;
						len += w;
						if(w<1) return len;
						dsp --;
					}
					// Append digits
					w = wtr(ptr,buf+i,CHARS_STRBUF-i);
					if(w<0) return w;
					len += w;
					if(w<(CHARS_STRBUF-i)) return len;
					// Append padding
					dsp -= CHARS_STRBUF-i;
					while(dsp>0) {
						_Chars_size_t j = 0;
						while(j<CHARS_TMPBUF && j<dsp) tmp[j++] = pad;
						w = wtr(ptr,tmp,j);
						if(w<0) return w;
						len += w;
						if(w<j) return len;
						dsp -= j;
					}
				}
				else {			// Right align
					// Reserve space for sign
					if(v<0) dsp --;
					else if(v>0 && sign) dsp --;
					// Append padding
					dsp -= CHARS_STRBUF-i;
					while(dsp>0) {
						_Chars_size_t j = 0;
						while(j<CHARS_TMPBUF && j<dsp) tmp[j++] = pad;
						w = wtr(ptr,tmp,j);
						if(w<0) return w;
						len += w;
						if(w<j) return len;
						dsp -= j;
					}
					// Append sign
					if(v<0) { 
						_Chars_char_t sgn = '-';
						w = wtr(ptr,&sgn,1);
						if(w<0) return w;
						len += w;
						if(w<1) return len;
						dsp --;
					}
					else if(v>0 && sign && out<len) {
						_Chars_char_t sgn = '+';
						w = wtr(ptr,&sgn,1);
						if(w<0) return w;
						len += w;
						if(w<1) return len;
						dsp --;
					}
					// Append digits
					w = wtr(ptr,buf+i,CHARS_STRBUF-i);
					if(w<0) return w;
					len += w;
					if(w<(CHARS_STRBUF-i)) return len;
				}
				  } break;
			case 'u': {			// --- Decimal unsigned integer -----------
				unsigned long u = (unsigned long) va_arg(ap, int);
				_Chars_size_t i = 0;
				// Pad with ' ' by default
				if(pad=='\0') pad = ' ';
				// Base is 10 by default
				if(pre==0) pre = 10;
				i = CHARS_STRBUF;
				while(i>0 && u!=0) {
					buf[--i] = '0' + u%pre;
					u /= pre;
				}
				if(dsp==0) {		// Fast version if no format width
					w = wtr(ptr,buf+i,CHARS_STRBUF-i);
					if(w<0) return w;
					len += w;
					if(w<(CHARS_STRBUF-i)) return len;
				}
				else if(left) {		// Left align
					// Append digits
					w = wtr(ptr,buf+i,CHARS_STRBUF-i);
					if(w<0) return w;
					len += w;
					if(w<(CHARS_STRBUF-i)) return len;
					// Append padding
					dsp -= CHARS_STRBUF-i;
					while(dsp>0) {
						_Chars_size_t j = 0;
						while(j<CHARS_TMPBUF && j<dsp) tmp[j++] = pad;
						w = wtr(ptr,tmp,j);
						if(w<0) return w;
						len += w;
						if(w<j) return len;
						dsp -= j;
					}
				}
				else {			// Right align
					// Append padding
					dsp -= CHARS_STRBUF-i;
					while(dsp>0) {
						_Chars_size_t j = 0;
						while(j<CHARS_TMPBUF && j<dsp) tmp[j++] = pad;
						w = wtr(ptr,tmp,j);
						if(w<0) return w;
						len += w;
						if(w<j) return len;
						dsp -= j;
					}
					// Append digits
					w = wtr(ptr,buf+i,CHARS_STRBUF-i);
					if(w<0) return w;
					len += w;
					if(w<(CHARS_STRBUF-i)) return len;
				}
			}	break;
			case 'x': case 'X': {		// --- Hexadecimal unsigned integer -------
				unsigned long u = (unsigned long) va_arg(ap, int);
				_Chars_size_t i = 0;
				// Pad with ' ' by default
				if(pad=='\0') pad = ' ';
				// Shift is 4 by default
				if(pre==0) pre = 4;
				if(pre>=6) pre = 6;
				unsigned char msk = (1<<pre)-1;
				if(chr=='X') chr = 'A';
				else chr = 'a';
				i = CHARS_STRBUF;
				while(i>0 && u!=0) {
					unsigned char d = u&msk;
					if(d<10) buf[--i] = '0' + d;
					else buf[--i] = chr + (d-10);
					u >>= pre;
				}
				if(left) {
					if(trnc && (CHARS_STRBUF-i)>dsp) {
						// Append higher order digits
						w = wtr(ptr,buf+i,dsp);
						if(w<0) return w;
						len += w;
						if(w<dsp) return len;
					}
					else {
						// Append digits
						w = wtr(ptr,buf+i,CHARS_STRBUF-i);
						if(w<0) return w;
						len += w;
						if(w<(CHARS_STRBUF-i)) return len;
						// Append padding
						dsp -= CHARS_STRBUF-i;
						while(dsp>0) {
							_Chars_size_t j = 0;
							while(j<CHARS_TMPBUF && j<dsp) tmp[j++] = pad;
							w = wtr(ptr,tmp,j);
							if(w<0) return w;
							len += w;
							if(w<j) return len;
							dsp -= j;
						}
					}
				}
				else {
					if(trnc && (CHARS_STRBUF-i)>dsp) {
						// Append lower order digits
						w = wtr(ptr,buf+CHARS_STRBUF-dsp,dsp);
						if(w<0) return w;
						len += w;
						if(w<dsp) return len;
					}
					else {
						// Append padding
						dsp -= CHARS_STRBUF-i;
						while(dsp>0) {
							_Chars_size_t j = 0;
							while(j<CHARS_TMPBUF && j<dsp) tmp[j++] = pad;
							w = wtr(ptr,tmp,j);
							if(w<0) return w;
							len += w;
							if(w<j) return len;
							dsp -= j;
						}
						// Append digits
						w = wtr(ptr,buf+i,CHARS_STRBUF-i);
						if(w<0) return w;
						len += w;
						if(w<(CHARS_STRBUF-i)) return len;
					}
				}
			}	break;
			default: {			// --- Other format types: ignored --------
			}
		} // End format character switch
	} // End Item loop
	// ### Return number of characters written ################################################
	return len;
}


_Chars_size_t _Chars_m(wtrnfmt)(_Chars_size_t (* wtr)(void * ptr, void const * buf, size_t len), void * ptr, _Chars_char_t const * fmt, ...) {
	va_list ap;
	_Chars_size_t res;
	va_start(ap,fmt);
	res = _Chars_v(wtrnfmt)(wtr,ptr,fmt,ap);
	va_end(ap);
	return res;
}



