
/*################################################################################################
 *##################                                                            ##################
 *##################                     Indexed Hash Set                       ##################
 *##################                                                            ##################
 *##################                      Implementation                        ##################
 *##################                                                            ##################
 *################################################################################################
 */

/*
 * ISet: High-efficiency Indexed Hash Set, using double open indexing
 *
 * Copyright (C) 2006-2007 Yann LANDRIN-SCHWEITZER, a.k.a. Varkhan
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:  Yann LANDRIN-SCHWEITZER, a.k.a. Varkhan
 * Contact: varkhan@varkhan.net
 * Website: http://www.varkhan.net/
 * Project: http://www.varkhan.net/software/VCom
 */



#include <malloc.h>
#include "VCom/Bits.h"

// Include header
#include "_ISet_.h"


//#define DEBUG

/*################################################################################################
 *###
 *###  Structures and types
 *###
 */


struct _ISet_struct_t {
	_ISet_mix_t		mixf;	// The mixing function
	_ISet_cmp_t		cmpf;	// The comparator function
	_ISet_size_t		size;	// The number of occupied entries in the slot table
	_ISet_size_t		head;	// The upper limit of used indexes
	_ISet_size_t		free;	// The number of available entries in the slot table
	_ISet_size_t		capa;	// The capacity of the slot table (size + free + deleted = capa), and primary hashing factor
	_ISet_size_t		mixr;	// The mixing (secondary hashing) factor
	_ISet_size_t		bpow;	// The 2-logarithm of the block size (block size is 2^bpow)
	float			lfct;	// The proportion of capa/size to enforce
	float			gfct;	// The table size growth factor
	_ISet_idx_t * *		hidx;	// A pointer to the index table
	void * *		hkey;	// A pointer to the key table
};



/*################################################################################################
 *###
 *###  Access macros
 *###
 */



#define ISET_DLOPEN_MIN_CAPA	8


#define ISET_DLOPEN_BLK_LEN(b)		((_ISet_size_t)((b)*sizeof(_ISet_key_t)+((b)>>3)))
#define ISET_DLOPEN_BLK_BITS(p,b)	( (unsigned char*) (p) )
#define ISET_DLOPEN_BLK_VALS(p,b)	( (_ISet_key_t*) ( ((void *)(p)) + ((b)>>3) ) )
#define ISET_DLOPEN_BLK_GETBIT(p,o)	( ((unsigned char *)(p))[(o)>>3] &   ((1<<((o)&0x7))) )
#define ISET_DLOPEN_BLK_SETBIT(p,o)	  ((unsigned char *)(p))[(o)>>3] |=  ((1<<((o)&0x7)))
#define ISET_DLOPEN_BLK_DELBIT(p,o)	  ((unsigned char *)(p))[(o)>>3] &= ~((1<<((o)&0x7)))
#define ISET_DLOPEN_BLK_CLR(p,b)	{ \
		unsigned long _zi = 0, _zn = (b)>>3, _zj; \
		while((_zj=_zi+sizeof(unsigned long))<=_zn) { \
			(*(unsigned long *)((p)+_zi)) = 0; \
			_zi = _zj; \
		} \
		while(_zi<_zn) (*(unsigned char *)((p)+(_zi++))) = 0; \
	}
#define ISET_DLOPEN_BLK_CPY(p,q,b)	{ \
		unsigned long _zi = 0, _zn = (b)>>3, _zj; \
		while((_zj=_zi+sizeof(unsigned long))<=_zn) { \
			(*(unsigned long *)((p)+_zi)) = (*(unsigned long *)((q)+_zi)); \
			_zi = _zj; \
		} \
		while(_zi<_zn) { \
			(*(unsigned char *)((p)+(_zi))) = (*(unsigned char *)((q)+(_zi))); \
			_zi++; \
		} \
	}



//#define DEBUG_ADJUST 3

#ifdef DEBUG_ADJUST
#define DEBUG
#endif

#ifdef DEBUG
int printf(const char *format, ...);
void _ISet_m(printr)(_ISet_obj_t iset, _ISet_size_t capa, _ISet_idx_t * * hidx, void ** hkey) {
	_ISet_size_t ibpos = 0;
	printf("ISet %lu %lu {\n",(unsigned long)iset->size,(unsigned long)iset->capa);
	if(hidx==NULL) { hidx = iset->hidx; capa = iset->capa; }
	if(hkey==NULL) { hkey = iset->hkey; capa = iset->capa; }
	if(capa==-1) capa = iset->capa;
	_ISet_size_t bpow = iset->bpow;
	_ISet_size_t mixr = iset->mixr;
	_ISet_size_t bnum = (capa==0)?0:((capa>>bpow)+1);
	_ISet_size_t blen = 1 << bpow;
	_ISet_size_t bmsk = blen - 1;
	while(ibpos<bnum) {
		_ISet_size_t iboff = 0;
		_ISet_idx_t * iblock = hidx[iboff];
		if(iblock==NULL) while(iboff<blen && ((ibpos<<bpow)+iboff)<capa) {
			printf("\t[%05ld] %c%c ----\n",(ibpos<<bpow)+iboff,' ',' ');
			iboff++;
		}		
		else while(iboff<blen && ((ibpos<<bpow)+iboff)<capa) {
			_ISet_idx_t idx = iblock[iboff];
			if(idx>0) {
				idx --;
				_ISet_size_t kbpos = idx >> bpow;
				_ISet_size_t kboff = idx & bmsk;
				void * kblock = hkey[kbpos];
				if(kblock==NULL || !ISET_DLOPEN_BLK_GETBIT(kblock,kboff)) {
					printf("\t[%05ld] %c%c %05lu -> X\n",(ibpos<<bpow)+iboff,'+','#',idx);
				}
				else {
					_ISet_key_t key = ISET_DLOPEN_BLK_VALS(kblock,blen)[kboff];
					_ISet_size_t hash = iset->mixf(key);
					printf("\t[%05ld] %c%c %05lu -> %05lu (%05lu/%05lu)\n",(ibpos<<bpow)+iboff,'+','*',idx,(unsigned long)key,hash%capa,(hash%mixr)+1);
				}
			}
			else if(idx<0) {
				idx = -idx-1;
				_ISet_size_t kbpos = idx >> bpow;
				_ISet_size_t kboff = idx & bmsk;
				void * kblock = hkey[kbpos];
				if(kblock==NULL || !ISET_DLOPEN_BLK_GETBIT(kblock,kboff)) {
					printf("\t[%05ld] %c%c %05lu -> X\n",(ibpos<<bpow)+iboff,' ','#',idx);
				}
				else {
					_ISet_key_t key = ISET_DLOPEN_BLK_VALS(kblock,blen)[kboff];
					_ISet_size_t hash = iset->mixf(key);
					printf("\t[%05ld] %c%c %05lu -> %05lu (%05lu/%05lu)\n",(ibpos<<bpow)+iboff,' ','*',idx,(unsigned long)key,hash%capa,(hash%mixr)+1);
				}
			}
			else printf("\t[%05ld] %c%c ----\n",(ibpos<<bpow)+iboff,' ',' ');
			iboff++;
		}
		ibpos++;
	}
	printf("}\n");
}
#endif



/*################################################################################################
 *###
 *###  Memory management
 *###
 */


//#if   ISET_DLOPEN_SIZE+0 ==  8
//_ISet_size_t _ISet_m(mixf)(unsigned char key);
//#elif ISET_DLOPEN_SIZE+0 == 16
//_ISet_size_t _ISet_m(mixf)(unsigned short key);
//#elif ISET_DLOPEN_SIZE+0 == 32
//_ISet_size_t _ISet_m(mixf)(unsigned int key);
//#elif ISET_DLOPEN_SIZE+0 == 64
//_ISet_size_t _ISet_m(mixf)(unsigned long key);
//#else
_ISet_size_t _ISet_m(mixf)(_ISet_key_t key);
//#endif


// Indexed Hash Set constructor
_ISet_obj_t _ISet_m(new)(_ISet_size_t blen, _ISet_size_t capa, _ISet_mix_t mixf, _ISet_cmp_t cmpf, float lfct, float gfct) {
	_ISet_obj_t iset;
	iset = malloc(sizeof(struct _ISet_struct_t));
#ifdef DEBUG_MEM
	printf("alloc(%lu) = %p\n",sizeof(struct _ISet_struct_t),list);
#endif
	if(iset==NULL) return NULL;
	if(mixf!=NULL) iset->mixf = mixf;
	else iset->mixf = _ISet_m(mixf);
	iset->cmpf = cmpf;
	_ISet_size_t bpow = vcom_llobit(blen);
	if(bpow<5) bpow = 5;
	iset->bpow = bpow;
	if(lfct>=1) lfct = .75f;
	iset->lfct = lfct;
	if(gfct<=1) gfct = 1.5f;
	iset->gfct = gfct;
	iset->capa = 0;
	iset->size = 0;
	iset->free = 0;
	iset->head = 0;
	_ISet_m(adj)(iset,capa);
//	if(capa==((_ISet_size_t)-1) || capa<ISET_DLOPEN_BLK_MIN_SIZE) capa = ISET_DLOPEN_BLK_MIN_SIZE;
//	capa = ISET_DLOPEN_COMP_CAPA(capa);
//	iset->capa = capa;
//	iset->mixr = _ISet_m(mixr)_(capa);
//	_ISet_size_t bnum = ((capa)>>bpow)+1;
//	_ISet_idx_t * * hidx = malloc(bnum*sizeof(_ISet_idx_t *));
//	if(iset->hidx==NULL) { free(iset); return NULL; }
//	_ISet_key_t * * hkey = malloc(bnum*sizeof(_ISet_key_t *));
//	if(hkey==NULL) { free(iset); return NULL; }
//	_ISet_size_t bpos = 0;
//	while(bpos<bnum) {
//		hidx[bpos] = 0;
//		hkey[bpos] = NULL;
//		bpos ++;
//	}
//	iset->hidx = hidx;
//	iset->hkey = hkey;
	return iset;
}

// Indexed Hash Set duplicator
_ISet_obj_t _ISet_m(cpy)(_ISet_obj_t iset) {
	_ISet_obj_t ciset;
	ciset = malloc(sizeof(struct _ISet_struct_t));
	if(ciset==NULL) return NULL;
	ciset->mixf = iset->mixf;
	ciset->cmpf = iset->cmpf;
	ciset->bpow = iset->bpow;
	ciset->lfct = iset->lfct;
	ciset->gfct = iset->gfct;
	ciset->capa = iset->capa;
	ciset->mixr = iset->mixr;
	_ISet_size_t bnum = ((iset->capa) >> (iset->bpow))+1;
	_ISet_size_t blen = 1 << (iset->bpow);
	_ISet_idx_t * * hidx = malloc(bnum*sizeof(_ISet_idx_t *));
	if(iset->hidx==NULL) { free(iset); return NULL; }
	void * * hkey = malloc(bnum*sizeof(void *));
	if(hkey==NULL) { free(iset); return NULL; }
	_ISet_size_t bpos = 0;
	while(bpos<bnum) {
		_ISet_idx_t * iblock = iset->hidx[bpos];
		if(iblock==NULL) hidx[bpos] = NULL;
		else {
			_ISet_idx_t * ciblock = malloc(blen*sizeof(_ISet_idx_t));
			if(ciblock==NULL) {
				free(hidx);
				free(hkey);
				free(ciset);
				return NULL;
			}
			_ISet_size_t boff = 0;
			while(boff<blen) {
				ciblock[boff] = iblock[boff];
				boff++;
			}
			hidx[bpos] = ciblock;
		}
		void * kblock = iset->hkey[bpos];
		if(kblock==NULL) hkey[bpos] = NULL;
		else {
			void * ckblock = malloc(ISET_DLOPEN_BLK_LEN(blen));
			if(ckblock==NULL) {
				free(hidx);
				free(hkey);
				free(ciset);
				return NULL;
			}
			ISET_DLOPEN_BLK_CPY(ckblock,kblock,blen);
			hkey[bpos] = ckblock;
		}
		bpos ++;
	}
	ciset->hidx = iset->hidx;
	ciset->hkey = iset->hkey;
	ciset->size = iset->size;
	ciset->free = iset->free;
	ciset->head = iset->head;
	return ciset;
}

// Indexed Hash Set destructor
void _ISet_m(del)(_ISet_obj_t iset) {
	if(iset==NULL) return;
	_ISet_size_t bnum = ((iset->capa)>>(iset->bpow))+1;
	_ISet_idx_t * * hidx = iset->hidx;
	void * * hkey = iset->hkey;
	if(hidx!=NULL) {	
		_ISet_size_t bpos = 0;
		while(bpos<bnum) {
			if(hidx[bpos]!=NULL) free(hidx[bpos]);
			bpos++;
		}
		free(hidx);
		iset->hidx = NULL;
	}
	if(hkey!=NULL) {
		_ISet_size_t bpos = 0;
		while(bpos<bnum) {
			if(hkey[bpos]!=NULL) free(hkey[bpos]);
			bpos++;
		}
		free(hkey);
		iset->hkey = NULL;
	}
	iset->mixf = NULL;
	iset->capa = 0;
	iset->size = 0;
	iset->free = 0;
	free(iset);
}


/*################################################################################################
 *###
 *###  Accessors and adjustment
 *###
 */


// Indexed Hash Set size
_ISet_size_t _ISet_m(size)(_ISet_obj_t iset) {
	if(iset==NULL) return -1;
	return iset->size;
}

// Indexed Hash Set size
_ISet_idx_t _ISet_m(head)(_ISet_obj_t iset) {
	if(iset==NULL) return -1;
	return iset->head;
}

// Indexed Hash Set capacity
_ISet_size_t _ISet_m(capa)(_ISet_obj_t iset) {
	if(iset==NULL) return -1;
	return iset->capa;
}


/*########################################
 *##  Global information accessors
 */


void _ISet_m(clr)(_ISet_obj_t iset) {
	if(iset==NULL) return;
	_ISet_size_t bnum = ((iset->capa)>>iset->bpow)+1;
	_ISet_size_t capa = iset->capa;
	if(iset->free==capa) return;
	iset->free = capa;
	iset->size = 0;
	iset->head = 0;
	_ISet_idx_t * * hidx = iset->hidx;
	void * * hkey = iset->hkey;
	if(hidx!=NULL) {	
		_ISet_size_t bpos = 0;
		while(bpos<bnum) {
			if(hidx[bpos]!=NULL) {
				free(hidx[bpos]);
				hidx[bpos] = NULL;
			}
			bpos++;
		}
	}
	if(hkey!=NULL) {
		_ISet_size_t bpos = 0;
		while(bpos<bnum) {
			if(hkey[bpos]!=NULL) {
				free(hkey[bpos]);
				hkey[bpos] = NULL;
			}
			bpos++;
		}
	}
}


/*########################################
 *##  Set elements accessors
 */


_ISet_idx_t _ISet_m(idx)(_ISet_obj_t iset, _ISet_key_t key) {
	if(iset==NULL) return -1;
	_ISet_size_t capa = iset->capa;
	_ISet_size_t mixr = iset->mixr;
	_ISet_size_t hash = iset->mixf(key);
	_ISet_size_t bpow = iset->bpow;
	_ISet_size_t blen = 1 << bpow;
	_ISet_size_t bmsk = blen - 1;
	// The slot position, starting at the primary hash
	_ISet_size_t slot = hash % capa;
	// The signed lookup index in the key table
//	signed long sidx = ISET_DLOPEN_GET_IDX(iset,slot);
	_ISet_size_t ibpos = slot >> bpow;
	_ISet_idx_t * iblock = iset->hidx[ibpos];
	if(iblock==NULL) return -1;
	_ISet_size_t iboff = slot & bmsk;
	signed long sidx = iblock[iboff];
	_ISet_key_t kval;
	_ISet_size_t kbpos;
	_ISet_size_t kboff;
	void * kblock;
	if(sidx==0) return -1;
	if(sidx>0) {
		sidx--;
		kbpos = sidx >> bpow;
		kblock = iset->hkey[kbpos];
		if(kblock==NULL) return -1;
		kboff = sidx & bmsk;
		kval = ISET_DLOPEN_BLK_VALS(kblock,blen)[kboff];
		// Key found
		if( (iset->mixf(kval)==hash) && ((iset->cmpf==NULL)?(key==kval):!(iset->cmpf(key,kval))) ) return sidx;
	}
	// The scan increment
	_ISet_size_t scan = ( hash % mixr ) + 1;
	while(1) {
		slot += scan;
		if(slot>=capa) slot -= capa;
		ibpos = slot >> bpow;
		iboff = slot & bmsk;
		iblock = iset->hidx[ibpos];
		if(iblock==NULL) return -1;
		sidx = iblock[iboff];
		// There's always an EMPTY slot
		if(sidx==0) return -1;
		if(sidx>0) {
			sidx--;
			kbpos = sidx >> bpow;
			kblock = iset->hkey[kbpos];
			if(kblock==NULL) return -1;
			kboff = sidx & bmsk;
			kval = ISET_DLOPEN_BLK_VALS(kblock,blen)[kboff];
			// Key found
			if( (iset->mixf(kval)==hash) && ((iset->cmpf==NULL)?(key==kval):!(iset->cmpf(key,kval))) ) return sidx;
		}
	}
}

_ISet_bool_t _ISet_m(has)(_ISet_obj_t iset, _ISet_idx_t idx) {
	if(iset==NULL || idx<0 || idx>=iset->head) return 0;
//	return ISET_DLOPEN_HAS_KEY(iset,idx);
	_ISet_size_t bpow = iset->bpow;
	_ISet_size_t blen = 1 << bpow;
	_ISet_size_t kbpos = idx >> bpow;
	void * kblock = iset->hkey[kbpos];
	if(kblock==NULL) return 0;
	_ISet_size_t kboff = idx & (blen-1);
	return ISET_DLOPEN_BLK_GETBIT(kblock,kboff);
}

_ISet_key_t _ISet_m(get)(_ISet_obj_t iset, _ISet_idx_t idx) {
    	if(iset==NULL || idx<0 || idx>=iset->head) return NULL;
//    	if(!ISET_DLOPEN_HAS_KEY(iset,idx)) return NULL;
//	return ISET_DLOPEN_GET_KEY(iset,idx);
	_ISet_size_t bpow = iset->bpow;
	_ISet_size_t blen = 1 << bpow;
	_ISet_size_t kbpos = idx >> bpow;
	void * kblock = iset->hkey[kbpos];
	if(kblock==NULL) return NULL;
	_ISet_size_t kboff = idx & (blen-1);
	if(!ISET_DLOPEN_BLK_GETBIT(kblock,kboff)) return NULL;
	return ISET_DLOPEN_BLK_VALS(kblock,blen)[kboff];
}

_ISet_idx_t _ISet_m(put)(_ISet_obj_t iset, _ISet_key_t key) {
	if(iset==NULL) return -1;
	_ISet_size_t capa = iset->capa;
	_ISet_size_t mixr = iset->mixr;
	_ISet_size_t hash = iset->mixf(key);
	_ISet_size_t bpow = iset->bpow;
	_ISet_size_t blen = (1<<bpow);
	_ISet_size_t bmsk = blen-1;
	// The slot position, starting at the primary hash
	_ISet_size_t slot = hash % capa;
	// The signed lookup index in the key table
//	signed long sidx = ISET_DLOPEN_GET_IDX(iset,slot);
	_ISet_size_t ibpos = slot >> bpow;
	_ISet_size_t iboff = slot & bmsk;
	_ISet_idx_t * iblock = iset->hidx[ibpos];
	signed long sidx = (iblock==NULL)?0:iblock[iboff];
#ifdef DEBUG
	printf("Put %lu: %lu (%lu [%lu,%lu] +%lu) = %ld\n",(unsigned long)key,hash,slot,ibpos,iboff,(hash%mixr)+1,(signed long)sidx);
#endif
	_ISet_key_t kval;
	_ISet_size_t kbpos;
	_ISet_size_t kboff;
	void * kblock;
	if(sidx==0) {
		sidx = iset->head;
		// Mark slot as OCCUPIED
		if(iblock==NULL) {
			iblock = malloc(blen*sizeof(_ISet_idx_t));
			if(iblock==NULL) return -1;
			iset->hidx[ibpos] = iblock;
			_ISet_size_t bpos = 0;
			while(bpos<blen) {
				iblock[bpos] = 0;
				bpos ++;
			}
		}
		kbpos = sidx >> bpow;
		kblock = iset->hkey[kbpos];
		if(kblock==NULL) {
			kblock = malloc(ISET_DLOPEN_BLK_LEN(blen));
			if(kblock==NULL) return -1;
			iset->hkey[kbpos] = kblock;
			ISET_DLOPEN_BLK_CLR(kblock,blen);
		}
		iblock[iboff] = sidx+1;
		iset->head++;
		iset->free--;
		kboff = sidx & bmsk;
		ISET_DLOPEN_BLK_VALS(kblock,blen)[kboff] = key;
		ISET_DLOPEN_BLK_SETBIT(kblock,kboff);
		// Too many elements for the capacity? rehash at an increased capa
		if((++iset->size)>=(iset->capa*iset->lfct)) _ISet_m(adj)(iset, (_ISet_size_t)(iset->size * iset->gfct) );
		// Too many deletions? rehash at the same capa for cleanup
		if(iset->free==0) _ISet_m(adj)(iset, -1);
		return sidx;
	}
	if(sidx>0) {
		sidx--;
		kbpos = sidx >> bpow;
		kblock = iset->hkey[kbpos];
		if(kblock==NULL) return -1;
		kboff = sidx & bmsk;
		kval = ISET_DLOPEN_BLK_VALS(kblock,blen)[kboff];
		// Key found
		if( (iset->mixf(kval)==hash) && ((iset->cmpf==NULL)?(key==kval):!(iset->cmpf(key,kval))) ) return sidx;
	}
	// The scan increment
	_ISet_size_t scan = ( hash % mixr ) + 1;
#ifdef DEBUG
	printf("Put %lu: starting scan at %lu by %lu\n",(unsigned long)key,slot,scan);
#endif
	while(1) {
		slot += scan;
		if(slot>=capa) slot -= capa;
		ibpos = slot >> bpow;
		iboff = slot & bmsk;
		iblock = iset->hidx[ibpos];
		sidx = (iblock==NULL)?0:iblock[iboff];
#ifdef DEBUG
	printf("Put %lu: %lu (%lu [%lu,%lu] +%lu) = %ld\n",(unsigned long)key,hash,slot,ibpos,iboff,(hash%mixr)+1,(signed long)sidx);
#endif
		// We just found an empty slot
		if(sidx==0) {
			sidx = iset->head;
			iset->free--;
			// Mark slot as OCCUPIED
			if(iblock==NULL) {
				iblock = malloc(blen*sizeof(_ISet_idx_t));
				if(iblock==NULL) return -1;
				iset->hidx[ibpos] = iblock;
				_ISet_size_t bpos = 0;
				while(bpos<blen) {
					iblock[bpos] = 0;
					bpos ++;
				}
			}
			kbpos = sidx >> bpow;
			kblock = iset->hkey[kbpos];
			if(kblock==NULL) {
				kblock = malloc(ISET_DLOPEN_BLK_LEN(blen));
				if(kblock==NULL) return -1;
				iset->hkey[kbpos] = kblock;
				ISET_DLOPEN_BLK_CLR(kblock,blen);
			}
			iblock[iboff] = sidx+1;
			iset->head++;
			iset->free--;
			kboff = sidx & bmsk;
			ISET_DLOPEN_BLK_VALS(kblock,blen)[kboff] = key;
			ISET_DLOPEN_BLK_SETBIT(kblock,kboff);
			// Too many elements for the capacity? rehash at an increased capa
			if((++iset->size)>=(iset->capa*iset->lfct)) _ISet_m(adj)(iset, (_ISet_size_t)(iset->size * iset->gfct) );
			// Too many deletions? rehash at the same capa for cleanup
			if(iset->free==0) _ISet_m(adj)(iset, -1);
			return sidx;
		}
		if(sidx>0) {
			sidx--;
			kbpos = sidx >> bpow;
			kblock = iset->hkey[kbpos];
			if(kblock==NULL) return -1;
			kboff = sidx & bmsk;
			kval = ISET_DLOPEN_BLK_VALS(kblock,blen)[kboff];
			// Key found
			if( (iset->mixf(kval)==hash) && ((iset->cmpf==NULL)?(key==kval):!(iset->cmpf(key,kval))) ) return sidx;
			continue;
		}
		break;
	}
	// If we get there, we encountered a REMOVED slot... look for an eventual match after that
	// Copy all slot values in temp vars
	signed long dsidx = -sidx;
	_ISet_size_t diboff = iboff;
	_ISet_idx_t * diblock = iblock;
	// We are on a REMOVED spot, so we have to continue scanning for the key
	while(1) {
		slot += scan;
		if(slot>=capa) slot -= capa;
		ibpos = slot >> bpow;
		iboff = slot & bmsk;
		iblock = iset->hidx[ibpos];
		sidx = (iblock==NULL)?0:iblock[iboff];
		if(sidx==0) break;
		if(sidx>0) {
			sidx--;
			kbpos = sidx >> bpow;
			kblock = iset->hkey[kbpos];
			if(kblock==NULL) return -1;
			kboff = sidx & bmsk;
			kval = ISET_DLOPEN_BLK_VALS(kblock,blen)[kboff];
			// Key found
			if( (iset->mixf(kval)==hash) && ((iset->cmpf==NULL)?(key==kval):!(iset->cmpf(key,kval))) ) return sidx;
		}
	}
	// Use saved DELETED slot for insertion -- block is necessarily not NULL
	// Mark as OCCUPIED, set index value as exact
	diblock[diboff] = dsidx--;
	kbpos = sidx >> bpow;
	kblock = iset->hkey[kbpos];
	if(kblock==NULL) {
		kblock = malloc(ISET_DLOPEN_BLK_LEN(blen));
		if(kblock==NULL) return -1;
		iset->hkey[kbpos] = kblock;
		ISET_DLOPEN_BLK_CLR(kblock,blen);
	}
	kboff = sidx & bmsk;
	ISET_DLOPEN_BLK_VALS(kblock,blen)[kboff] = key;
	ISET_DLOPEN_BLK_SETBIT(kblock,kboff);
	// Too many elements for the capacity? rehash at an increased capa
	if((++iset->size)>=(iset->capa*iset->lfct)) _ISet_m(adj)(iset, (_ISet_size_t)(iset->size * iset->gfct) );
	// Too many deletions? rehash at the same capa for cleanup
	if(iset->free==0) _ISet_m(adj)(iset, -1);
	return sidx;
}

_ISet_key_t * _ISet_m(rem)(_ISet_obj_t iset, _ISet_idx_t idx, _ISet_key_t * key) {
    	if(iset==NULL || idx>=iset->head) return NULL;
	_ISet_size_t bpow = iset->bpow;
	_ISet_size_t blen = (1<<bpow);
	_ISet_size_t bmsk = blen-1;
	_ISet_size_t kbpos = idx >> bpow;
	_ISet_size_t kboff = idx & bmsk;
	void * kblock = iset->hkey[kbpos];
	if(kblock==NULL) return NULL;
	if(!ISET_DLOPEN_BLK_GETBIT(kblock,kboff)) return NULL;
    	// We need to look up the key, to provide a starting point for the slot list
    	_ISet_key_t kval = ISET_DLOPEN_BLK_VALS(kblock,blen)[kboff];
//	ISET_DLOPEN_BLK_VALS(kblock,blen)[kboff] = NULL;
    	_ISet_size_t capa = iset->capa;
	_ISet_size_t hash = iset->mixf(kval);
	_ISet_size_t mixr = iset->mixr;
	// The slot position, starting at the primary hash
	_ISet_size_t slot = hash % capa;
	_ISet_size_t ibpos = slot >> bpow;
	_ISet_size_t iboff = slot & bmsk;
	_ISet_idx_t * iblock = iset->hidx[ibpos];
	if(iblock==NULL) {
		// Something is wrong with the hash if we get here
		return NULL;
	}
	signed long sidx = iblock[iboff];
	// The actual lookup index in the key table
	signed long pidx = idx+1;
	signed long nidx = -pidx;
	if ( sidx!=0 && sidx!=pidx && sidx!=nidx ) {
		// The scan increment
		_ISet_size_t scan = ( hash % mixr ) + 1;
		do {
			slot += scan;
			if(slot>=capa) slot -= capa;
			ibpos = slot >> bpow;
			iboff = slot & bmsk;
			iblock = iset->hidx[ibpos];
			if(iblock==NULL) {
				// Something is wrong with the hash if we get here
				return NULL;
			}
			sidx = iblock[iboff];
		} while( sidx!=0 && sidx!=pidx && sidx!=nidx );
		// At this point, we have either an EMPTY slot, or an OCCUPIED or DELETED slot with the right index
	}
#ifdef DEBUG
	printf("Rem %lu -> %lu (%ld) at %lu (%lu,%lu) \n",(unsigned long)kval,idx,(signed long)sidx,slot,ibpos,iboff);
#endif
	// If EMPTY or DELETED, nothing to be done
	if(sidx<=0) return NULL;
	// At this point we necessarily have an OCCUPIED slot, and sidx == idx+1
	ISET_DLOPEN_BLK_DELBIT(kblock,kboff);
	iblock[iboff] = nidx;
	iset->size--;
	if(key!=NULL) *key = kval;
	return key;
}


static _ISet_size_t _ISet_m(prime)(_ISet_size_t capa);

void _ISet_m(adj)(_ISet_obj_t iset, _ISet_size_t capa) {
	if(iset==NULL) return;
	_ISet_size_t mixr = iset->mixr;
	if(capa==-1) {
		capa = iset->capa;
		mixr = iset->mixr;
	}
	else {
		if(capa==0||capa<iset->size*iset->lfct) {
			if(capa<ISET_DLOPEN_MIN_CAPA) capa = ISET_DLOPEN_MIN_CAPA;
			if(capa<iset->size) capa = iset->size;
		}
		capa = _ISet_m(prime)((_ISet_size_t)(capa/iset->lfct)+1);
		mixr = capa-2;
	}
	_ISet_size_t bpow = iset->bpow;
	_ISet_size_t blen = 1<<bpow;
	_ISet_size_t bmsk = blen-1;
	_ISet_size_t bnum = (iset->capa==0)?0:((iset->capa>>bpow)+1);
	_ISet_size_t nbnum = (capa>>bpow)+1;
#ifdef DEBUG
	printf("Adj %lu (%lu) -> %lu (%lu)\n",iset->capa,bnum,capa,nbnum);
#endif
	_ISet_idx_t * * hidx = malloc(nbnum*sizeof(_ISet_idx_t *));
	if(hidx==NULL) { return; }
//	_ISet_key_t * * hkey = malloc(nbnum*sizeof(_ISet_key_t *));
//	if(hkey==NULL) { return; }
	_ISet_size_t bpos = 0;
	while(bpos<nbnum) {
		hidx[bpos] = NULL;
//		hkey[bpos] = NULL;
		bpos ++;
	}
	_ISet_size_t head = 0;
	_ISet_size_t ibpos = 0;
	while(ibpos<bnum) {
		_ISet_idx_t * iblock = iset->hidx[ibpos];
		if(iblock==NULL) { ibpos++; continue; }
		_ISet_size_t iboff = 0;
		while(iboff<blen) {
			_ISet_idx_t idx = iblock[iboff];
			if(idx<=0) { iboff++; continue; }
			_ISet_size_t kbpos = (idx-1) >> bpow;
			_ISet_size_t kboff = (idx-1) & bmsk; 
			void * kblock = iset->hkey[kbpos];
			if(kblock==NULL) { iboff++; continue; }
			if(!ISET_DLOPEN_BLK_GETBIT(kblock,kboff)) { iboff++; continue; }
			if(head<idx) head = idx;
			_ISet_key_t kval = ISET_DLOPEN_BLK_VALS(kblock,blen)[kboff];
			_ISet_size_t hash = iset->mixf(kval);
			// The slot position, starting at the primary hash
			_ISet_size_t slot = hash % capa;
			_ISet_size_t nibpos = slot >> bpow;
			_ISet_size_t niboff = slot & bmsk;
			_ISet_idx_t * niblock = hidx[nibpos];
#ifdef DEBUG2
			printf("Move idx %lu from %lu,%lu to %lu,%lu\n",idx-1,ibpos,iboff,nibpos,niboff);
#endif
			if(niblock!=NULL && niblock[niboff]!=0) {
				_ISet_size_t scan = (hash%mixr)+1;
				do {
					slot += scan;
					if(slot>=capa) slot -= capa;
					nibpos = slot >> bpow;
					niboff = slot & bmsk;
					niblock = hidx[nibpos];
				} while(niblock!=NULL && niblock[niboff]!=0);
			}
			if(niblock==NULL) {
				niblock = malloc(blen*sizeof(_ISet_idx_t));
				if(niblock==NULL) return;
				hidx[nibpos] = niblock;
				_ISet_size_t bpos = 0;
				while(bpos<blen) {
					niblock[bpos] = 0;
					bpos ++;
				}
			}
			niblock[niboff] = idx;
			iboff++;
		}
		ibpos++;
	}
	// If we just can't fit the existing indexes, rehash with enough room
	if(capa<=head) {
		free(hidx);
		_ISet_m(adj)(iset,head);
		return;
	}
	// Change the size of the keys if needed
	if(nbnum!=bnum) {
		void * * hkey = realloc(iset->hkey,nbnum*sizeof(void *));
		if(hkey==NULL) return;
		while(bnum<nbnum) {
			hkey[bnum++] = NULL;
		}
		iset->hkey = hkey;
	}
	iset->capa = capa;
	iset->mixr = mixr;
	iset->free = iset->capa - iset->size;
	iset->head = head;
	iset->hidx = hidx;
}



/*################################################################################################
 *###
 *###  Mixing and prime functions
 *###
 */

static long PRIMES_LENGTH = 464;
static long PRIMES_VALUES[] = {
	3,		3,		3,		3,		3,		3,		3,		3,
	3,		3,		5,		5,		5,		5,		5,		5,
	7,		7,		7,		7,		7,		7,		7,		7,
	7,		7,		7,		13,		13,		13,		13,		13,
	13,		13,		13,		19,		19,		19,		19,		19,
	19,		19,		19,		19,		19,		19,		19,		31,
	31,		31,		31,		31,		31,		31,		43,		43,
	43,		43,		43,		43,		43,		43,		61,		61,
	61,		61,		61,		73,		73,		73,		73,		73,
	73,		73,		103,		103,		109,		109,		109,		109,
	109,		139,		139,		151,		151,		151,		151,		181,
	181,		193,		199,		199,		199,		229,		241,		241,
	241,		271,		283,		283,		313,		313,		313,		349,
	349,		349,		349,		421,		433,		463,		463,		463,
	523,		523,		571,		601,		619,		661,		661,		661,
	661,		661,		823,		859,		883,		883,		883,		1021,
	1063,		1093,		1153,		1153,		1231,		1321,		1321,		1429,
	1489,		1489,		1621,		1699,		1789,		1873,		1951,		2029,
	2131,		2143,		2311,		2383,		2383,		2593,		2731,		2803,
	3001,		3121,		3259,		3391,		3583,		3673,		3919,		4093,
	4273,		4423,		4651,		4801,		5023,		5281,		5521,		5743,
	5881,		6301,		6571,		6871,		7129,		7489,		7759,		8089,
	8539,		8863,		9283,		9721,		10141,		10531,		11071,		11551,
	12073,		12613,		13009,		13759,		14323,		14869,		15649,		16363,
	17029,		17839,		18541,		19471,		20233,		21193,		22159,		23059,
	24181,		25171,		26263,		27541,		28753,		30013,		31321,		32719,
	34213,		35731,		37309,		38923,		40639,		42463,		44281,		46309,
	48313,		50461,		52711,		55051,		57529,		60091,		62299,		65521,
	68281,		71413,		74611,		77713,		81373,		84979,		88663,		92671,
	96739,		100801,		105529,		109849,		115021,		120079,		125509,		131011,
	136861,		142873,		149251,		155863,		162751,		169891,		177433,		185071,
	193381,		202129,		211063,		220021,		229981,		240349,		250969,		262111,
	273643,		285841,		298411,		311713,		325543,		339841,		355009,		370663,
	386989,		404269,		422113,		440809,		460081,		480463,		501829,		524221,
	547399,		571603,		596929,		623353,		651019,		679909,		709741,		741343,
	774133,		808441,		844201,		881539,		920743,		961531,		1004119,	1048573,
	1094923,	1143283,	1193911,	1246963,	1302181,	1359733,	1420039,	1482853,
	1548541,	1616899,	1688413,	1763431,	1841293,	1922773,	2008081,	2097133,
	2189989,	2286883,	2388163,	2493853,	2604013,	2719669,	2840041,	2965603,
	3097123,	3234241,	3377191,	3526933,	3682363,	3845983,	4016041,	4193803,
	4379719,	4573873,	4776223,	4987891,	5208523,	5439223,	5680153,	5931313,
	6194191,	6468463,	6754879,	7053331,	7366069,	7692343,	8032639,	8388451,
	8759953,	9147661,	9552733,	9975193,	10417291,	10878619,	11360203,	11863153,
	12387841,	12936529,	13509343,	14107801,	14732413,	15384673,	16065559,	16777141,
	17519893,	18295633,	19105483,	19951231,	20834689,	21757291,	22720591,	23726449,
	24776953,	25873963,	27018853,	28215619,	29464579,	30769093,	32131711,	33554011,
	35039911,	36591211,	38211163,	39903121,	41669479,	43514521,	45441199,	47452879,
	49553941,	51747991,	54039079,	56431513,	58930021,	61539091,	64263571,	67108669,
	70079959,	73182409,	76422793,	79806229,	83339383,	87029053,	90881083,	94906249,
	99108043,	103495879,	108077731,	112863013,	117860053,	123078019,	128526943,	134217439,
	140159911,	146365159,	152845393,	159612601,	166679173,	174058849,	181765093,	189812341,
	198216103,	206991601,	216156043,	225726379,	235720159,	246156271,	257054491,	268435009,
	280319203,	292730833,	305691181,	319225021,	333358513,	348117151,	363529759,	379624279,
	396432481,	413983771,	432312511,	451452613,	471440161,	492312523,	514109251,	536870839,
	560640001,	585461743,	611382451,	638450569,	666717199,	696235363,	727060069,	759249643,
	792864871,	827967631,	864625033,	902905501,	942880663,	984625531,	1028218189,	1073741719,
	1121280091,	1170923713,	1222764841,	1276901371,	1333434301,	1392470281,	1454120779,	1518500173,
	1585729993,	1655935399,	1729249999,	1805811253,	1885761133,	1969251079,	2056437379,	2147482951
};


static _ISet_size_t _ISet_m(prime)(_ISet_size_t capa) {
	_ISet_size_t min = 0;
	_ISet_size_t max = PRIMES_LENGTH-1;
	while(min<=max) {
		// Warning: risk of overflow here if table is too big
		_ISet_size_t mid = (min+max)>>1;
		_ISet_size_t val = PRIMES_VALUES[mid];
		if(val<capa) min = mid+1;
		else if(val>capa) max = mid-1;
		else return val;
	}
	if(min>=PRIMES_LENGTH) return PRIMES_VALUES[PRIMES_LENGTH-1];
	return PRIMES_VALUES[min];
}


#if   ISET_DLOPEN_SIZE+0 ==  8

_ISet_size_t _ISet_m(mixf)(_ISet_key_t key) {
	unsigned long val = (unsigned long) key;
	return val ^ ( ( val + ((~val)>>3) ) << 5 ) ;
}

#elif ISET_DLOPEN_SIZE+0 == 16

_ISet_size_t _ISet_m(mixf)(_ISet_key_t key) {
        return 2654435761LL*(unsigned long)key ;
}

#elif ISET_DLOPEN_SIZE+0 == 32
// Jenkin's 32-bit hash function
_ISet_size_t _ISet_m(mixf)(_ISet_key_t key) {
	unsigned long val = (unsigned long) key;
	val += val << 12 ;
	val ^= val >> 22 ;
	val += val <<  4 ;
	val ^= val >>  9 ;
	val += val << 10 ;
	val ^= val >>  2 ;
	val += val <<  7 ;
	val ^= val >> 12 ;
	return val ;
}

#elif ISET_DLOPEN_SIZE+0 == 64

_ISet_size_t _ISet_m(mixf)(_ISet_key_t key) {
	unsigned long val = (unsigned long) key;
	val = (~val) + (val << 21);
	val ^= val >> 25 ;
	val += val <<  6 ;
	val ^= val >> 13 ;
	val += val <<  2 ;
	val += val <<  4 ;
	val ^= val >> 29 ;
	val += val <<  3 ;
	val += val << 28;
	return val;
}

#else
// Jenkin's 32-bit hash function
_ISet_size_t _ISet_m(mixf)(_ISet_key_t key) {
	unsigned long val = (unsigned long) key;
	val = (~val) + (val << 21);
	val ^= val >> 25 ;
	val += val <<  6 ;
	val ^= val >> 13 ;
	val += val <<  2 ;
	val += val <<  4 ;
	val ^= val >> 29 ;
	val += val <<  3 ;
	val += val << 28;
	return val;
}
#endif


