//################################################################################################
//##################                                                            ##################
//##################                    Indexed Sparse List                     ##################
//##################                                                            ##################
//##################                      Implementation                        ##################
//##################                                                            ##################
//################################################################################################


/*
 * ListSparse: High-efficiency Indexed List based on a sparse bucket array.
 *
 * Copyright (C) 2006-2009 Yann LANDRIN-SCHWEITZER, a.k.a. Varkhan
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:  Yann LANDRIN-SCHWEITZER, a.k.a. Varkhan
 * Contact: varkhan@varkhan.net
 * Website: http://www.varkhan.net/
 * Project: http://www.varkhan.net/software/VCom
 */


#include <malloc.h>
#include "VCom/Bits.h"

// Include header
#include "_List_.h"

//#define DEBUG
//#define DEBUG_MEM

#ifdef DEBUG_MEM
#ifndef DEBUG
#define DEBUG
#endif
#endif


/*################################################################################################
 *###
 *###  Structures and types
 *###
 */


// The List structure
struct _List_struct_t {
	_List_size_t	size;	// The number of elements
	_List_size_t	head;	// The upper bound of indexes
	_List_size_t	bnum;	// The number of allocated blocks
	_List_size_t	bpow;	// The 2-logarithm of the block size (block size is 2^bpow)
	void * *	heap;	// The storage blocks
};



/*################################################################################################
 *###
 *###  Access macros
 *###
 */


#define LIST_SPARSE_BLK_MIN_SIZE	8
#define LIST_SPARSE_BLK_LEN(b,s)	((_List_size_t)((s)*sizeof(_List_val_t)+((b)>>3)))
#define LIST_SPARSE_BLK_SIZE(p,b)	( vcom_vctbit((void *)(p),(b)>>3) )
#define LIST_SPARSE_BLK_HEAD(p,b)	( vcom_vhibit((void *)(p),(b)>>3) )
#define LIST_SPARSE_BLK_TAIL(p,b)	( vcom_vlobit1((void *)(p),(b)>>3) )
#define LIST_SPARSE_BLK_CLEAR(p,b)	{ \
		unsigned long _zi = 0, _zn = (b)>>3, _zj; \
		while((_zj=_zi+sizeof(unsigned long))<=_zn) { \
			(*(unsigned long *)((p)+_zi)) = 0; \
			_zi = _zj; \
		} \
		while(_zi<_zn) (*(unsigned char *)((p)+(_zi++))) = 0; \
	}
#define LIST_SPARSE_BLK_COPY(p,c,l)	{ \
		unsigned long _ci = 0, _cj; \
		while((_cj=_ci+sizeof(unsigned long))<=(l)) { \
			(*(unsigned long *)((c)+_ci)) = (*(unsigned long *)((p)+_ci)); \
			_ci = _cj; \
		} \
		while(_ci<(l)) { \
			(*(unsigned char *)((c)+_ci)) = (*(unsigned char *)((p)+_ci)); \
			_ci ++; \
		} \
	}
#define LIST_SPARSE_BLK_BITS(p,b)	( (unsigned char*) (p) )
#define LIST_SPARSE_BLK_VALS(p,b)	( (_List_val_t*) ( ((void *)(p)) + ((b)>>3) ) )
#define LIST_SPARSE_BLK_GETBIT(p,o)	( ((unsigned char *)(p))[(o)>>3] &   ((1<<((o)&0x7))) )
#define LIST_SPARSE_BLK_SETBIT(p,o)	  ((unsigned char *)(p))[(o)>>3] |=  ((1<<((o)&0x7)))
#define LIST_SPARSE_BLK_DELBIT(p,o)	  ((unsigned char *)(p))[(o)>>3] &= ~((1<<((o)&0x7)))
#define LIST_SPARSE_BLK_GETPOS(p,o)	( vcom_vctbit((void *)(p),(o)>>3) + vcom_bctbit( ((unsigned char *)(p))[(o)>>3] & (((1<<((o)&0x7))-1)) ) )

#ifdef DEBUG
#include <stdio.h>
#define printb(p,b)	{							\
	_List_size_t _pi = 0;							\
	while(_pi<((b)>>5)) {							\
		unsigned long _px = ((unsigned long*)(p))[_pi++];		\
		printf("%lX%lX%lX%lX%lX%lX%lX%lX",				\
				_px&0xF,(_px&0xF0)>>4,				\
				(_px&0xF00)>>8, (_px&0xF000)>>12,		\
				(_px&0xF0000)>>16, (_px&0xF00000)>>20,		\
				(_px&0xF000000)>>24, (_px&0xF0000000)>>28	\
				);						\
	}									\
}
#endif




/*################################################################################################
 *###
 *###  Object lifecycle
 *###
 */


// List constructor
_List_obj_t _List_m(new)(_List_size_t blen, _List_size_t capa) {
	_List_size_t bpow = vcom_llobit(blen);
	if(bpow<5) bpow = 5;
	if(capa==((_List_size_t)-1) || capa<LIST_SPARSE_BLK_MIN_SIZE) capa = LIST_SPARSE_BLK_MIN_SIZE;
	_List_size_t bnum = ((capa-1)>>bpow)+1;
	_List_obj_t list;
	list = malloc(sizeof(struct _List_struct_t));
#ifdef DEBUG_MEM
	printf("alloc(%lu) = %p\n",sizeof(struct _List_struct_t),list);
#endif
	if(list==NULL) return NULL;
	void * * heap = malloc(bnum*sizeof(void *));
#ifdef DEBUG_MEM
	printf("alloc(%p->heap,%lu) = %p\n",list,bnum*sizeof(void *),heap);
#endif
	if(heap==NULL) { free(list); return NULL; }
	_List_size_t bpos = 0;
	while(bpos<bnum) heap[bpos++] = NULL;
	list->heap = heap;
	list->bpow = bpow;
	list->bnum = bnum;
	list->head = 0;
	list->size = 0;
	return list;
}

// List duplicator
_List_obj_t _List_m(cpy)(_List_obj_t list) {
	_List_obj_t copy;
	copy = malloc(sizeof(struct _List_struct_t));
#ifdef DEBUG_MEM
	printf("alloc(%lu) = %p\n",sizeof(struct _List_struct_t),copy);
#endif
	if(copy==NULL) return NULL;
	_List_size_t blen = 1<<(list->bpow);
	_List_size_t bnum = list->bnum;
	void * * heap = malloc(bnum*sizeof(void *));
#ifdef DEBUG_MEM
	printf("alloc(%p->heap,%lu) = %p\n",copy,bnum*sizeof(void *),heap);
#endif
	if(heap==NULL) { free(copy); return NULL; }
	_List_size_t bpos = 0;
	while(bpos<bnum) {
		void * block = list->heap[bpos];
		if(block!=NULL) {
			_List_size_t bcnt = LIST_SPARSE_BLK_LEN(blen,LIST_SPARSE_BLK_SIZE(block,blen));
			heap[bpos] = malloc(bcnt);
#ifdef DEBUG_MEM
			printf("alloc(%p->heap[%lu],%lu) = %p)\n",copy,bpos,bcnt,heap[bpos]);
#endif
			LIST_SPARSE_BLK_COPY(block,heap[bpos],bcnt);
		}
		else {
			heap[bpos] = NULL;
		}
		bpos ++;
	}
	copy->heap = heap;
	copy->bpow = list->bpow;
	copy->bnum = list->bnum;
	copy->head = list->head;
	copy->size = list->size;
	return copy;
}


// List destructor
void _List_m(del)(_List_obj_t list) {
	if(list==NULL) return;
	if(list->heap!=NULL) {
		_List_size_t bnum = list->bnum;
		_List_size_t bpos = 0;
		while(bpos<bnum) {
#ifdef DEBUG_MEM
			printf("free(%p->heap[%lu] = %p)\n",list,bpos,list->heap[bpos]);
#endif
			free(list->heap[bpos]);
			list->heap[bpos] = NULL;
			bpos ++;
		}
		free(list->heap);
#ifdef DEBUG_MEM
		printf("free(%p->heap = %p)\n",list,list->heap);
#endif
	}
	list->bnum = 0;
	list->size = 0;
	list->head = 0;
	free(list);
#ifdef DEBUG_MEM
	printf("free(%p)\n",list);
#endif
}




/*################################################################################################
 *###
 *###  Parameters accessors
 *###
 */


// List size
_List_size_t _List_m(size)(_List_obj_t list) {
	if(list==NULL) return -1;
	return list->size;
}

// List head
_List_idx_t _List_m(head)(_List_obj_t list) {
	if(list==NULL) return -1;
	return list->head;
}

_List_idx_t _List_m(tail)(_List_obj_t list) {
	if(list==NULL) return -1;
	_List_size_t bpos = 0;
	_List_size_t blen = 1<<list->bpow;
	while(bpos<list->bnum) {
		void * block = list->heap[bpos];
		if(block==NULL) return bpos<<list->bpow;
		_List_size_t boff = LIST_SPARSE_BLK_TAIL(block,blen);
		if(boff<blen) return (bpos<<list->bpow)+boff;
		bpos ++;
	}
	return list->head;
}


// List capacity
_List_size_t _List_m(capa)(_List_obj_t list) {
	if(list==NULL) return -1;
	return list->bnum<<list->bpow;
}

// List block length
_List_size_t _List_m(blen)(_List_obj_t list) {
	if(list==NULL) return -1;
	return 1<<(list->bpow);
}



/*################################################################################################
 *###
 *###  Bulk operations
 *###
 */


// Clears all elements
void _List_m(clr)(_List_obj_t list) {
	if(list==NULL) return;
	_List_size_t bpos = 0;
	_List_size_t bnum = list->bnum;
	void * * heap = list->heap;
	while(bpos<bnum) heap[bpos++] = NULL;
	list->head = 0;
	list->size = 0;
}


// List adjustment
void _List_m(adj)(_List_obj_t list, _List_size_t capa) {
	if(list==NULL) return;
	if(capa==((_List_size_t)-1) || capa<LIST_SPARSE_BLK_MIN_SIZE) capa = LIST_SPARSE_BLK_MIN_SIZE;
	if(capa<=list->size) return;
	_List_size_t bnum = ((capa-1)>>list->bpow)+1;
#ifdef DEBUG_MEM
	printf("adjust(%p->heap,%lu,%lu)\n",list,list->bnum,bnum);
#endif
	void * * heap = realloc(list->heap,bnum*sizeof(void *));
#ifdef DEBUG_MEM
	printf("realloc(%p->heap = %p,%lu) = %p\n",list,list->heap,bnum*sizeof(void *),heap);
#endif
	if(heap==NULL) return;
	_List_size_t bpos = list->bnum;
	while(bpos<bnum) heap[bpos++] = NULL;
	list->heap = heap;
	list->bnum = bnum;
}



/*################################################################################################
 *###
 *###  Data interface
 *###
 */

// List test
_List_bool_t _List_m(has)(_List_obj_t list, _List_idx_t idx) {
	if(list==NULL) return 0;
	if(list->heap==NULL || idx>=list->head) return 0;
	_List_size_t bpos = idx >> list->bpow;
	void * block = list->heap[bpos];
	if(block==NULL) return 0;
	_List_size_t blen = 1<<list->bpow;
	_List_size_t boff = idx & (blen-1);
	return 0!=LIST_SPARSE_BLK_GETBIT(block,boff);
}


// List search
_List_val_t * _List_m(get)(_List_obj_t list, _List_idx_t idx) {
	if(list==NULL) return NULL;
	if(list->heap==NULL || idx>=list->head) return NULL;
	_List_size_t bpos = idx >> list->bpow;
	void * block = list->heap[bpos];
	if(block==NULL) return NULL;
	_List_size_t blen = 1<<list->bpow;
	_List_size_t boff = idx & (blen-1);
	if(!LIST_SPARSE_BLK_GETBIT(block,boff)) return NULL;
	return LIST_SPARSE_BLK_VALS(block,blen) + LIST_SPARSE_BLK_GETPOS(block,boff);
}


// List update
_List_val_t * _List_m(set)(_List_obj_t list, _List_idx_t idx, _List_val_t val) {
	if(list==NULL) return NULL;
	_List_size_t bpos = idx >> list->bpow;
	if(list->heap==NULL || bpos>=list->bnum) {
		_List_m(adj)(list,idx+1);
	}
	_List_size_t blen = 1<<list->bpow;
	_List_size_t boff = idx & (blen-1);
	void * block = list->heap[bpos];
	if(block==NULL) {
		block = malloc(LIST_SPARSE_BLK_LEN(blen,1));
#ifdef DEBUG_MEM
		printf("alloc(%p->heap[%lu],%lu) = %p\n",list,bpos,LIST_SPARSE_BLK_LEN(blen,1),block);
#endif
		if(block==NULL) return NULL;
		LIST_SPARSE_BLK_CLEAR(block,blen);
#ifdef DEBUG_BIT
		printf("Set bit %p->bits[%lu] = %p <- %lu\n",list,bpos,LIST_SPARSE_BLK_BITS(block,blen),boff);
#endif
		LIST_SPARSE_BLK_SETBIT(block,boff);
#ifdef DEBUG_BIT
		printf("Set val %p->vals[%lu] = %p <- %lu (%lu)\n",list,bpos,LIST_SPARSE_BLK_VALS(block,blen),LIST_SPARSE_BLK_GETPOS(block,boff),LIST_SPARSE_BLK_SIZE(block,blen));
#endif
		_List_val_t * pval = LIST_SPARSE_BLK_VALS(block,blen) + LIST_SPARSE_BLK_GETPOS(block,boff);
		*pval = val;
		list->heap[bpos] = block;
		if(idx>=list->head) list->head = idx+1;
		list->size ++;
		return pval;
	}
	if(LIST_SPARSE_BLK_GETBIT(block,boff)) {
		_List_val_t * pval = LIST_SPARSE_BLK_VALS(block,blen) + LIST_SPARSE_BLK_GETPOS(block,boff);
		*pval = val;
		return pval;
	}
	_List_size_t n = LIST_SPARSE_BLK_SIZE(block,blen);
	block = realloc(block,LIST_SPARSE_BLK_LEN(blen,n+1));
#ifdef DEBUG_MEM
	printf("realloc(%p->heap[%lu] = %p,%lu) = %p\n",list,bpos,list->heap[bpos],LIST_SPARSE_BLK_LEN(blen,n+1),block);
#endif
	if(block==NULL) return NULL;
	_List_size_t p = LIST_SPARSE_BLK_GETPOS(block,boff);
#ifdef DEBUG_BIT
	printf("Set bit %p->bits[%lu] = %p <- %lu\n",list,bpos,LIST_SPARSE_BLK_BITS(block,blen),boff);
#endif
	LIST_SPARSE_BLK_SETBIT(block,boff);
#ifdef DEBUG_BIT
	printf("Set val %p vals[%lu] = %p <- %lu (%lu)\n",list,bpos,LIST_SPARSE_BLK_VALS(block,blen),LIST_SPARSE_BLK_GETPOS(block,boff),LIST_SPARSE_BLK_SIZE(block,blen));
#endif
	_List_val_t * pval = LIST_SPARSE_BLK_VALS(block,blen);
	while(n>p) {
		pval[n] = pval[n-1];
		n--;
	}
	pval += p;
	*pval = val;
	list->heap[bpos] = block;
	if(idx>=list->head) list->head = idx+1;
	list->size ++;
	return pval;
}


// List delete
_List_val_t * _List_m(rem)(_List_obj_t list, _List_idx_t idx, _List_val_t * val) {
	if(list==NULL) return NULL;
//	if(list->heap==NULL || idx>=list->head) return NULL;
	_List_size_t bpos = idx >> list->bpow;
	_List_size_t blen = 1<<list->bpow;
	_List_size_t boff = idx & (blen-1);
//	printf("REM %lu (0x%lx/0x%lx) %lu (0x%lx/0x%lx) (block 9791 has %lu elts)\n",bpos,bpos,list->bnum,boff,boff,blen,list->bnum<=9791?0:list->heap[9791]==NULL?0:LIST_SPARSE_BLK_SIZE(list->heap[9791],blen));
//	printf("REM %lu: %lu (0x%lx/0x%lx) %lu (0x%lx/0x%lx)\n",idx,bpos,bpos,list->bnum,boff,boff,blen);
	if(bpos>=list->bnum) return NULL;
	void * block = list->heap[bpos];
	if(block==NULL) return NULL;
//	_List_size_t blen = 1<<list->bpow;
//	_List_size_t boff = idx & (blen-1);
	if(!LIST_SPARSE_BLK_GETBIT(block,boff)) return NULL;
	_List_size_t n = LIST_SPARSE_BLK_SIZE(block,blen);
	_List_size_t p = LIST_SPARSE_BLK_GETPOS(block,boff);
	_List_val_t * pval = LIST_SPARSE_BLK_VALS(block,blen);
	if(val!=NULL) *val = pval[p];
	n --;
	if(n==0) {
		LIST_SPARSE_BLK_DELBIT(block,boff);
		list->size --;
#ifdef DEBUG_MEM
		printf("free(%p->heap[%lu] = %p)\n",list,bpos,list->heap[bpos]);
#endif
		free(block);
		list->heap[bpos] = NULL;

		// We are likely to have changed the value of head, recompute
		bpos = 1+((list->head-1)>>list->bpow);
		if(bpos>list->bnum) bpos = list->bnum;
		while(bpos>0) {
			if(bpos<list->bnum && list->heap[bpos-1]!=NULL) break;
			bpos --;
		}
		// We end up with bpos just after the last non-NULL block
		if(bpos>0) {
			bpos --;
			// Now recompute according to the head of the current block
			block = list->heap[bpos];
			list->head = (bpos<<list->bpow) + LIST_SPARSE_BLK_HEAD(block,blen);
		}
		else list->head = 0;
		return val;
	}
	while(p<n) {
		p++;
		pval[p-1] = pval[p];
	}
	LIST_SPARSE_BLK_DELBIT(block,boff);
	list->size --;
	block = realloc(block,LIST_SPARSE_BLK_LEN(blen,n));
#ifdef DEBUG_MEM
	printf("realloc(%p->heap[%lu] = %p,%lu) = %p\n",list,bpos,list->heap[bpos],LIST_SPARSE_BLK_LEN(blen,n),block);
#endif
	if(block==NULL) return NULL;
	list->heap[bpos] = block;
	if((idx+1)==list->head) {
		// We are likely to have changed the value of head, recompute according to the head of the current block
		list->head = (bpos<<list->bpow) + LIST_SPARSE_BLK_HEAD(block,blen);
	}
	return val;
}

// Apply callback to all entries, in slot order
_List_size_t _List_m(visit)(_List_obj_t list, _List_vis_t visitor, void * dat) {
	if(list==NULL || list->heap==NULL) return -1;
	_List_size_t bnum = list->bnum;
	_List_size_t blen = 1<<list->bpow;
	_List_size_t rsum = 0;
	_List_size_t bpos = 0;
	while(bpos<bnum) {
		void * block = list->heap[bpos];
		if(block!=NULL) {
			_List_size_t boff = 0;
			_List_idx_t idx = bpos<<list->bpow;
			while(boff<blen) {
				if(LIST_SPARSE_BLK_GETBIT(block,boff)) {
					_List_val_t * pval = LIST_SPARSE_BLK_VALS(block,blen) + LIST_SPARSE_BLK_GETPOS(block,boff);
					_List_size_t rval = visitor(idx,pval,dat);
					if(rval==-1) return rsum;
					rsum += rval;
				}
				boff ++;
				idx ++;
			}
		}
		bpos ++;
	}
	return rsum;
}


#ifdef DEBUG
#include <stdio.h>
void _List_m(print)(_List_obj_t list) {
	if(list==NULL) {
		printf("@null\n");
		return;
	}
	printf("@%p (%lu/%lu) {\n",list,list->size,list->bnum<<list->bpow);
	_List_size_t blen = 1<<(list->bpow);
	_List_size_t bnum = list->bnum;
	_List_size_t bpos = 0;
	while(bpos<bnum) {
		void * block = list->heap[bpos];
		if(block!=NULL) {
			printf("\t@%p\t",block);
			printb(LIST_SPARSE_BLK_BITS(block,blen),blen);
			printf("\t");
			_List_val_t * pval = LIST_SPARSE_BLK_VALS(block,blen);
			_List_size_t n = LIST_SPARSE_BLK_SIZE(block,blen);
			_List_size_t p = 0;
			while(p<n) {
				printf("%p ",pval[p++]);
			}
			printf("\n");
		}
		else {
			printf("\t--\n");
		}
		bpos ++;
	}
	printf("}\n");
}
#endif


/*###
 *###  END
 *###
 *################################################################################################
 */




